import numpy as np

from matrx.actions.action import Action, ActionResult
from matrx.objects.standard_objects import Door


class OpenDoorAction(Action):
    """ Action that opens doors. """

    def __init__(self, duration_in_ticks=0):
        """ Action that opens doors.

        The action that opens a specific door within a given range from the agent.

        Parameters
        ----------
        duration_in_ticks : int
            Optional, default: ``1``. Should be zero or larger.

            The default duration of this action in ticks during which the
            :class:`matrx.grid_world.GridWorld` blocks the
            :class:`matrx.agents.agent.Agent` performing other actions. By default
            this is 1, meaning that this action  will take both the tick in which
            it was decided upon and the subsequent tick.

        """

        super().__init__(duration_in_ticks)

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Opens a door in the world.

        Mutates the `door_status` of an
        :class:`matrx.objects.simple_objects.Door` in the
        :class:`matrx.grid_world.GridWorld` to be open (and as such passable).
        It does so only if the `object_id` exists, is a
        :class:`matrx.objects.simple_objects.Door`, and is within
        `door_range`.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            door is sought according to the `object_id` parameter.
        agent_id : str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id : st
            Optional. Default: ``None``

            The string representing the unique identifier of the door that
            should be opened.
        door_range : int
            Optional. Default: ``np.inf``

            The maximum allowed distance between the agent and the door for
            the agent to be able to open that door.

        Returns
        -------
        OpenDoorActionResult
            The ActionResult depicting the action's success or failure and
            reason for that result.

            See :class:`matrx.actions.simple_objects.OpenDoorActionResult` for
            the results it can return.

        """

        # fetch options
        door_range = 1 if 'door_range' not in kwargs else kwargs['door_range']
        # object_id is required
        object_id = None if 'object_id' not in kwargs else kwargs['object_id']
        if object_id == None:
            result = OpenDoorActionResult(OpenDoorActionResult.NO_OBJECT_SPECIFIED, False)
            return result

        # get obj
        obj = grid_world.environment_objects[object_id]

        # call the open door action in the object
        obj.open_door()

        result = OpenDoorActionResult(OpenDoorActionResult.RESULT_SUCCESS, True)
        return result

    def is_possible(self, grid_world, agent_id, **kwargs):
        """ Check if the OpenDoorAction is possible.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the door
             is sought according to the `object_id` parameter.
        agent_id : str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id : str
            Optional. Default: ``None``

            The string representing the unique identifier of the door that
            should be opened.
        door_range : int
            Optional. Default: ``np.inf``

            The maximum allowed distance between the agent and the door for
            the agent to be able to open that door.

        Returns
        -------
        OpenDoorActionResult
            Depicts the action's expected success or failure and reason for
            that result.

            See :class:`matrx.actions.simple_objects.OpenDoorActionResult` for
            the results it can return.

        """
        # fetch options
        door_range = np.inf if 'door_range' not in kwargs else kwargs['door_range']
        object_id = None if 'object_id' not in kwargs else kwargs['object_id']

        result = _is_possible_door_open_close(grid_world, agent_id, OpenDoorActionResult, object_id, door_range)
        return result


class CloseDoorAction(Action):
    """ Action that closes doors. """

    def __init__(self, duration_in_ticks=0):
        """ Action that closes doors.

        The action that closes a specific door within a given range from the agent.

        Parameters
        ----------
        duration_in_ticks : int
            Optional. Default: ``1``. Should be zero or larger.

            The default duration of this action in ticks during which the
            :class:`matrx.grid_world.GridWorld` blocks the agent performing
            other actions. By default this is 1, meaning that this action will take
            both the tick in which it was decided upon and the subsequent tick.

        """
        super().__init__(duration_in_ticks)

    def mutate(self, grid_world, agent_id, **kwargs):
        """ Closes a door in the world.

        Mutates the `door_status` of an
        :class:`matrx.objects.simple_objects.Door` in the
        :class:`matrx.grid_world.GridWorld` to be closed (and as such not
        passable). It does so only if the `object_id` exists, is a
        :class:`matrx.objects.simple_objects.Door`, and is within
        `door_range`.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the
            door is sought according to the `object_id` parameter.
        agent_id : str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id : st
            Optional. Default: ``None``

            The string representing the unique identifier of the door that
            should be closed.
        door_range : int
            Optional. Default: ``np.inf``

            The maximum allowed distance between the agent and the door for
            the agent to be able to close that door.

        Returns
        -------
        OpenDoorActionResult
            Depicts the action's success or failure and reason for that result.

            See :class:`matrx.actions.simple_objects.CloseDoorActionResult`
            for the results it can return.

        """

        # object_id is required (the is_possible method already ensured us that the object is within door_range)
        object_id = None if 'object_id' not in kwargs else kwargs['object_id']
        if object_id is None:
            result = CloseDoorActionResult(CloseDoorActionResult.NO_OBJECT_SPECIFIED, False)
            return result

        # get obj
        obj = grid_world.environment_objects[object_id]

        # call the close door action in the object
        obj.close_door()

        result = CloseDoorActionResult(CloseDoorActionResult.RESULT_SUCCESS, True)
        return result

    def is_possible(self, grid_world, agent_id, **kwargs):
        """Check if the CloseDoorAction is possible.

        Parameters
        ----------
        grid_world : GridWorld
            The :class:`matrx.grid_world.GridWorld` instance in which the door
             is sought according to the `object_id` parameter.
        agent_id : str
            The string representing the unique identifier that represents the
            agent performing this action.
        object_id : str
            Optional. Default: ``None``

            The string representing the unique identifier of the door that
            should be closed.
        door_range : int
            Optional. Default: ``np.inf``

            The maximum allowed distance between the agent and the door for
            that agent to be able to close that door.

        Returns
        -------
        OpenDoorActionResult
            Depicts the action's expected success or failure and reason for
            that result.

            See :class:`matrx.actions.door_actions.CloseDoorActionResult`
            for the results it can return.

        """
        # fetch options
        door_range = np.inf if 'door_range' not in kwargs else kwargs['door_range']
        object_id = None if 'object_id' not in kwargs else kwargs['object_id']

        result = _is_possible_door_open_close(grid_world, agent_id, CloseDoorActionResult, object_id, door_range)
        return result


class CloseDoorActionResult(ActionResult):
    RESULT_SUCCESS = "Door was succesfully closed."
    NO_DOORS_IN_RANGE = "No door found in range"
    NOT_IN_RANGE = "Specified door is not within range."
    NOT_A_DOOR = "CloseDoor action could not be performed, as object isn't a door"
    RESULT_UNKNOWN_OBJECT_TYPE = 'obj_id is no Agent and no Object, unknown what to do'
    DOOR_ALREADY_CLOSED = "Can't close door, door is already closed."
    DOOR_BLOCKED = "Can't close door, object or agent is blocking the door opening."
    NO_OBJECT_SPECIFIED = "No object_id of a door specified to close."

    def __init__(self, result, succeeded):
        """ ActionResult for the CloseDoorAction.

        The results uniquely for
        :class:`matrx.actions.action.door_actions.CloseDoorAction` are (as class
        constants):

        * RESULT_SUCCESS: When the action is a success.
        * NO_DOORS_IN_RANGE: When no door objects are within the specified range.
        * NOT_IN_RANGE: When the given `object_id` is not within the specified
          range.
        * NOT_A_DOOR: When the given `object_id` does not exist.
        * DOOR_ALREADY_CLOSED: When the door is already closed.
        * DOOR_BLOCKED: When another object is at the door's location.
        * NO_OBJECT_SPECIFIED: When `object_id` is not given.

        Parameters
        ----------
        result : str
            A string representing the reason for the (expected) success or fail of
            a :class:`matrx.actions.door_actions.CloseDoorAction`.
        succeeded : bool
            A boolean representing the (expected) success or fail of a
            :class:`matrx.actions.door_actions.CloseDoorAction`.

        See Also
        --------
        CloseDoorAction

        """
        super().__init__(result, succeeded)


class OpenDoorActionResult(ActionResult):
    RESULT_SUCCESS = "Door was successfully opened."
    NO_DOORS_IN_RANGE = "No door found in range"
    NOT_IN_RANGE = "Specified door is not within range."
    NOT_A_DOOR = "OpenDoor action could not be performed, as object isn't a door"
    RESULT_UNKNOWN_OBJECT_TYPE = 'obj_id is no Agent and no Object, unknown what to do'
    DOOR_ALREADY_OPEN = "Can't open door, door is already open"
    DOOR_BLOCKED = "Can't close door, object or agent is blocking the door opening."
    NO_OBJECT_SPECIFIED = "No object_id of a door specified to open."

    def __init__(self, result, succeeded):
        """ ActionResult for the OpenDoorAction.

        The results uniquely for
        :class:`matrx.actions.door_actions.OpenDoorAction` are (as class
        constants):

        * RESULT_SUCCESS: When the action is a success.
        * NO_DOORS_IN_RANGE: When no door objects are within the specified range.
        * NOT_IN_RANGE: When the given `object_id` is not within the specified
          range.
        * NOT_A_DOOR: When the given `object_id` does not exist.
        * DOOR_ALREADY_OPEN: When the door is already open.
        * NO_OBJECT_SPECIFIED: When `object_id` is not given.

        Parameters
        ----------
        result : str
            A string representing the reason for the (expected) success or fail of
            an :class:`matrx.actions.door_actions.OpenDoorAction`.
        succeeded : bool
            A boolean representing the (expected) success or fail of an
            :class:`matrx.actions.door_actions.OpenDoorAction`.

        See Also
        --------
        OpenDoorAction

        """
        super().__init__(result, succeeded)


def _is_possible_door_open_close(grid_world, agent_id, action_result, object_id=None, door_range=np.inf):
    """ Private MATRX method.

    Checks if a door can be opened/closed based on the given `action_result`
    parameter.

    Parameters
    ----------
    grid_world : GridWorld
        The :class:`matrx.grid_world.GridWorld` instance in which the door is
         sought according to the `object_id` parameter.
    agent_id : str
        The string representing the unique identified that represents the agent
        performing this action.
    action_result : {Type[OpenDoorActionResult], Type[CloseDoorActionResult]}
        The type of the :class:`matrx.actions.action.ActionResult` that should
        be returned. Also used to determine for which kind of
        :class:`matrx.actions.action.Action` this check is made (
        :class:`matrx.actions.door_actions.OpenDoorAction` or
        :class:`matrx.actions.door_actions.CloseDoorAction`).
    object_id : str,
        Optional. Default: `None`

        The string representing the unique identifier of the door that should
        be opened.
    door_range : int
        Optional. Default: `np.inf`
        The maximum allowed distance between the agent and the door for the
        agent to be able to open or close that door.

    Returns
    -------
    ActionResult : {Type[OpenDoorActionResult], Type[CloseDoorActionResult]}
        Returns either an OpenDoorActionResult or CloseDoorActionResult
        (depends on the parameter action_result) that contains whether the
        respective action is possible or not.

    """
    reg_ag = grid_world.registered_agents[agent_id]  # Registered Agent
    loc_agent = reg_ag.location  # Agent location

    # check if there is a Door object in the scenario
    objects_in_range = grid_world.get_objects_in_range(loc_agent, object_type=Door, sense_range=door_range)

    # there is no Door in range, so not possible to open any door
    if len(objects_in_range) is 0:
        return action_result(action_result.NO_DOORS_IN_RANGE, False)

    # if we did not get a specific door to open, we simply return success as it is possible to open an arbitrary door
    # as there is atleast one door in range.
    if object_id is None:
        return action_result(action_result.RESULT_SUCCESS, True)

    # check if the given object_id even exists
    if object_id not in grid_world.environment_objects.keys():
        return action_result(action_result.NOT_A_DOOR, False)

    # check if the given object_id is an actual door in range
    if object_id not in objects_in_range.keys():
        return action_result(action_result.NOT_IN_RANGE, False)

    # get the target object
    obj = grid_world.environment_objects[object_id]

    # check if door is already open or closed
    if action_result == OpenDoorActionResult and obj.properties["is_open"]:
        return action_result(action_result.DOOR_ALREADY_OPEN, False)
    elif action_result == CloseDoorActionResult and not obj.properties["is_open"]:
        return action_result(action_result.DOOR_ALREADY_CLOSED, False)

    # when closing, check that there are no objects in the door opening
    if action_result == CloseDoorActionResult:
        # get all objects at the location of the door
        objects_in_door_opening = grid_world.get_objects_in_range(obj.location, object_type="*", sense_range=0)

        # more than 1 object at that location (besides the door itself) means the door is blocked
        if len(objects_in_door_opening) > 1:
            return action_result(action_result.DOOR_BLOCKED, False)

    return action_result(action_result.RESULT_SUCCESS, True)



import threading
import copy
import logging

import jsonpickle
from flask import Flask, jsonify, abort, request
from flask_cors import CORS

from matrx.messages.message import Message

'''
This file holds the code for the MATRX RESTful api.
External scripts can send POST and/or GET requests to retrieve state, tick and other information, and send
userinput or other information to MATRX. The api is a Flask (Python) webserver.

For visualization, see the seperate MATRX visualization folder / package.
'''

debug = True

app = Flask(__name__)
CORS(app)
port = 3001

# variables to be set by MATRX
# states is a list of length 'current_tick' with a dictionary containing all states of that tick, indexed by agent_id
states = []
current_tick = 0
tick_duration = 0.5
grid_size = [1, 1]
MATRX_info = {}
next_tick_info = {}
add_message_to_agent = None
received_messages = {}  # messages received via the api, intended for the Gridworld
gw_message_manager = None  # the message manager of the gridworld, containing all messages of various types
gw = None
teams = None  # dict with team names (keys) and IDs of agents who are in that team (values)
# currently only one world at a time is supported
current_world_ID = False

# a temporary state for the current tick, which will be written to states after all
# agents have been updated
temp_state = {}

# variables to be read (only!) by MATRX and set (only!) through api calls
userinput = {}
matrx_paused = False
matrx_done = False


#########################################################################
# api connection methods
#########################################################################


@app.route('/get_info', methods=['GET', 'POST'])
def get_info():
    """ Provides the general information on the world, contained in the world object.

    Returns
        MATRX world object, containing general information on the world and scenario.
    -------
    """
    MATRX_info['matrx_paused'] = matrx_paused
    return jsonify(MATRX_info)


@app.route('/get_latest_state_and_messages/<agent_id>', methods=['GET', 'POST'])
def get_latest_state_and_messages(agent_id):
    """ Provides all most recent information from MATRX: Both the state and messages from the latest
    tick for one particular agent, as well as the current MATRX status (paused or not).

    Parameters
    ----------
    agent_id
        The ID of the targeted agent

    Returns
    -------
        a dictionary containing the states under the "states" key, and the messages under the "messages" key.

    """
    # check for validity and return an error if not valid
    api_call_valid, error = check_states_API_request(ids=[agent_id])
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    # fetch states and messages
    states_ = __fetch_states(current_tick, agent_id)
    messages = gw_message_manager.fetch_messages(current_tick, current_tick, clean_input_ids(agent_id)[0])
    chatrooms = gw_message_manager.fetch_chatrooms(clean_input_ids(agent_id)[0])

    return jsonify({"matrx_paused": matrx_paused, "states": states_, "messages": messages, "chatrooms": chatrooms})


#########################################################################
# MATRX fetch state api calls
#########################################################################

@app.route('/get_states/<tick>', methods=['GET', 'POST'])
def get_states(tick):
    """ Provides the states of all agents (including the god view) from tick `tick` onwards to current tick.

    Parameters
    ----------
    tick
        integer indicating from which tick onwards to send the states.

    Returns
    -------
        Returns a list of length `tick` to current_tick. For each tick (item in the list), a dictionary contains the
        state for each agent existing in the simulation, indexed by their agent ID.

    """

    # check for validity and return an error if not valid
    api_call_valid, error = check_states_API_request(tick=tick)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    return jsonify(__fetch_states(tick))


@app.route('/get_states/<tick>/<agent_ids>', methods=['GET', 'POST'])
def get_states_specific_agents(tick, agent_ids):
    """ Provides the states starting from tick `tick` to current_tick, for the agents specified in `agent_ids`.

    Parameters
    ----------
    tick
        integer indicating from which tick onwards to send the states.
    agent_ids
        One agent ID, or a List of agent IDs for which the states should be returned. God view = "god"

    Returns
    -------
        Returns a list of length `tick` to current_tick. For each tick (item in the list), a dictionary contains the
        state for each agent as specified in `agent_ids`, indexed by their agent ID.

    """
    # check for validity and return an error if not valid
    api_call_valid, error = check_states_API_request(tick=tick)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    return jsonify(__fetch_states(tick, agent_ids))


@app.route('/get_latest_state/<agent_ids>', methods=['GET', 'POST'])
def get_latest_state(agent_ids):
    """ Provides the latest state of one or multiple agents

    Parameters
    ----------
    agent_ids
        IDs of agents for which to send the latest state. Either a single agent ID, or a list of agent IDs.
        God view = "god"

    Returns
    -------
        Returns a list of length `tick` to current_tick. For each tick, a dictionary contains the states for each
        agent as specified in `agent_ids`, indexed by their agent ID.

    """
    return get_states_specific_agents(current_tick, agent_ids)


@app.route('/get_filtered_latest_state/<agent_ids>', methods=['POST'])
def get_filtered_latest_state(agent_ids):
    """

    """

    # check for validity and return an error if not valid
    api_call_valid, error = check_states_API_request(tick=current_tick)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    # Get the agent states
    agent_states = __fetch_states(current_tick, agent_ids)[0]

    # Filter the agent states based on the received properties list
    props = request.json['properties']
    if 'filters' in request.json.keys():
        filters = request.json['filters']
    else:
        filters = None

    filtered_states = {}
    for agent_id, agent_dict in agent_states.items():
        state_dict = agent_dict['state']
        filtered_state_dict = __filter_dict(state_dict, props, filters)
        filtered_states[agent_id] = filtered_state_dict

    return jsonify(filtered_states)


#########################################################################
# MATRX fetch messages api calls
#########################################################################
@app.route('/get_messages/<tick>', methods=['GET', 'POST'])
def get_messages(tick):
    """ Provides the messages of all agents from tick `tick` onwards to current tick. Also returns
    the chatrooms at the latest tick.

    Parameters
    ----------
    tick
        integer indicating from which tick onwards to send the messages.

    Returns
    -------
        Returns a dictionary containing `messages` and `chatrooms`. The chatrooms subdictionary
        contains all accessible chatrooms at the latest tick. The messages subdictionary contains
        all messages for tick `tick` to the current tick,
        subdivided under `global`, `team`, and `private`.
        Also see the documentation of the
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_messages` and
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_chatrooms` functions.
    """

    # check for validity and return an error if not valid
    api_call_valid, error = check_messages_API_request(tick=tick)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    messages = gw_message_manager.fetch_messages(int(tick), current_tick)
    chatrooms = gw_message_manager.fetch_chatrooms()

    return jsonify({"messages": messages, "chatrooms": chatrooms})


@app.route('/get_messages/<tick>/<agent_id>', methods=['GET', 'POST'])
def get_messages_specific_agent(tick, agent_id):
    """ Provides all messages either send by or addressed to `agent_id`, from tick `tick` onwards.

    Parameters
    ----------
    tick
        integer indicating which ticks [`tick`,current_tick] (including `tick` and current_tick)
        the messages should be fetched from.
    agent_id
        The `agent_id` of the agent of whom the messages should be fetched.
        All fetched messages are either sent to or by the agent with agent ID `agent_id`.

    Returns
    -------
        Returns a dictionary containing `messages` and `chatrooms`. The chatrooms subdictionary
        contains all accessible chatrooms at the latest tick. The messages subdictionary contains
        all messages sent or received by `agent_id`,
        subdivided under `global`, `team`, and `private`.
        Also see the documentation of the
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_messages` and
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_chatrooms` functions.

    """
    # check for validity and return an error if not valid
    api_call_valid, error = check_messages_API_request(tick=tick, agent_id=agent_id)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    messages = gw_message_manager.fetch_messages(int(tick), current_tick, clean_input_ids(agent_id)[0])
    chatrooms = gw_message_manager.fetch_chatrooms(clean_input_ids(agent_id)[0])

    return jsonify({"messages": messages, "chatrooms": chatrooms})


@app.route('/get_latest_messages', methods=['GET', 'POST'])
def get_latest_messages():
    """ Provides all messages of the latest tick.

    Returns
        Returns a dictionary containing `messages` and `chatrooms`. The chatrooms subdictionary
        contains all accessible chatrooms at the latest tick. The messages subdictionary contains
        all messages of the latest tick, subdivided under `global`, `team`, and `private`.
        Also see the documentation of the
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_messages` and
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_chatrooms` functions.
    -------

    """
    # check for validity and return an error if not valid
    api_call_valid, error = check_messages_API_request(tick=current_tick)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    messages = gw_message_manager.fetch_messages(current_tick, current_tick)
    chatrooms = gw_message_manager.fetch_chatrooms()

    return jsonify({"messages": messages, "chatrooms": chatrooms})


@app.route('/get_latest_messages/<agent_id>', methods=['GET', 'POST'])
def get_latest_messages_specific_agent(agent_id):
    """ Provides the messages of the latest tick either sent by or addressed to `agent_id`.

    Parameters
    ----------
    agent_id
        The `agent_id` of the agent of whom the messages should be fetched.
        All fetched messages are either sent to or by the agent with agent ID `agent_id`.

    Returns
    -------
        Returns a dictionary containing `messages` and `chatrooms`. The chatrooms subdictionary
        contains all accessible chatrooms at the latest tick, set to or by the agent with
        ID `agent_id`. The messages subdictionary contains
        all messages of the latest tick, subdivided under `global`, `team`, and `private`.
        Also see the documentation of the
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_messages` and
        :func:`~matrx.utils.message_manager.MessageManager.MyClass.fetch_chatrooms` functions.

    """
    # check for validity and return an error if not valid
    api_call_valid, error = check_messages_API_request(tick=current_tick, agent_id=agent_id)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    messages = gw_message_manager.fetch_messages(current_tick, current_tick, clean_input_ids(agent_id)[0])
    chatrooms = gw_message_manager.fetch_chatrooms(clean_input_ids(agent_id)[0])

    return jsonify({"messages": messages, "chatrooms": chatrooms})


#########################################################################
# MATRX user input api calls
#########################################################################

@app.route('/send_userinput/<agent_ids>', methods=['POST'])
def send_userinput(agent_ids):
    """ Can be used to send user input from the user (pressed keys) to the specified human agent(s) in MATRX

    Parameters
    ----------
    agent_ids
        ID(s) of the human agent(s) to which the data should be passed.

    Returns
    -------
        returns True if the data was valid (right now always)
    -------

    """
    global userinput

    api_call_valid, error = check_states_API_request(current_tick, agent_ids, ids_required=True)
    if not api_call_valid:
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    # make sure the ids are a list
    try:
        agent_ids = eval(agent_ids)
    except:
        agent_ids = [agent_ids]

    # fetch the data from the request object
    data = request.json

    # add each pressed key as user input for each specified human agent
    for agent_id in agent_ids:
        for pressed_key in data:

            # add  the agent_id if not existing yet
            if agent_id not in userinput:
                userinput[agent_id] = []

            # save the pressed key of the agent
            userinput[agent_id].append(pressed_key)

    return jsonify(True)


@app.route('/send_message', methods=['POST'])
def send_message():
    """ Send a message containing information to one or multiple specific agent, the agent's team, or all agents.

    Message as defined in matrx.utils.message

    Returns
    -------
        Error if api call invalid, or True if valid.
    """
    # fetch the data
    data = request.json

    # check that all required parameters have been passed
    required_params = ("content", "sender", "receiver")
    if not all(k in data for k in required_params):
        error = {"error_code": 400, "error_message": f"Missing one of the required parameters: {required_params}"}
        print("api request not valid:", error)
        return abort(error['error_code'], description=error['error_message'])

    # create message
    msg = Message(content=data['content'], from_id=data['sender'], to_id=data['receiver'])

    # add the received_messages to the api global variable
    if data['sender'] not in received_messages:
        received_messages[data['sender']] = []
    received_messages[data['sender']].append(msg)

    return jsonify(True)


#########################################################################
# MATRX context menu API calls
#########################################################################
@app.route('/fetch_context_menu_of_self', methods=['POST'])
def fetch_context_menu_of_self():
    """ Fetch the context menu opened for a specific object/location of the agent being controlled by the user.
    """
    # fetch the data
    data = request.json

    # check that all required parameters have been passed
    required_params = ("agent_id_who_clicked", "clicked_object_id", "click_location", "self_selected")
    if not all(k in data for k in required_params):
        return return_error(code=400, message=f"Missing one of the required parameters: {required_params}")

    agent_id_who_clicked = data['agent_id_who_clicked']
    clicked_object_id = data['clicked_object_id']
    click_location = data['click_location']
    self_selected = data['self_selected']

    # check if agent_id_who_clicked exists in the gw
    if agent_id_who_clicked not in gw.registered_agents.keys() and agent_id_who_clicked != "god":
        return return_error(code=400, message=f"Agent with ID {agent_id_who_clicked} does not exist.")

    # ignore if called from the god view
    if agent_id_who_clicked.lower() == "god":
        return return_error(code=400,
                            message=f"The god view is not an agent and thus cannot show its own context menu.")

    # fetch context menu from agent
    context_menu = gw.registered_agents[agent_id_who_clicked].create_context_menu_for_self_func(clicked_object_id,
                                                                                                 click_location,
                                                                                                self_selected)

    # encode the object instance of the message
    for item in context_menu:
        item['Message'] = jsonpickle.encode(item['Message'])

    return jsonify(context_menu)



@app.route('/fetch_context_menu_of_other', methods=['POST'])
def fetch_context_menu_of_other():
    """ Fetch the context menu opened for a specific object/location of the agent being controlled by the user.
    """
    # fetch the data
    data = request.json

    # check that all required parameters have been passed
    required_params = ("agent_id_who_clicked", "clicked_object_id", "click_location", "agent_selected")
    if not all(k in data for k in required_params):
        return return_error(code=400, message=f"Missing one of the required parameters: {required_params}")

    agent_id_who_clicked = data['agent_id_who_clicked']
    clicked_object_id = data['clicked_object_id']
    click_location = data['click_location']
    agent_selected = data['agent_selected']

    # check if agent_id_who_clicked exists in the gw
    if agent_id_who_clicked not in gw.registered_agents.keys() and agent_id_who_clicked != "god":
        return return_error(code=400, message=f"Agent with ID {agent_id_who_clicked} does not exist.")

    # check if the selected agent exists
    if agent_selected not in gw.registered_agents.keys() :
        return return_error(code=400, message=f"Selected agent with ID {agent_selected} does not exist.")

    # ignore if called from the god view
    # if agent_id_who_clicked.lower() == "god":
    #     return return_error(code=400, message=f"The god view is not an agent and thus cannot show its own context menu.")

    # fetch context menu from agent
    context_menu = gw.registered_agents[agent_selected].create_context_menu_for_other_func(agent_id_who_clicked,
                                                                                                 clicked_object_id,
                                                                                                 click_location)

    # encode the object instance of the message
    for item in context_menu:
        item['Message'] = jsonpickle.encode(item['Message'])

    return jsonify(context_menu)



@app.route('/send_message_pickled', methods=['POST'])
def send_message_pickled():
    """ This function makes it possible to send a custom message to a MATRX agent via the API as a jsonpickle object.
    For instance, sending a custom message when a context menu option is clicked.
    The pre-formatted CustomMessage instance an be jsonpickled and sent via the API.
    This API call can handle that request and send the CustomMessage to the MATRX agent

        Returns
    -------
        Error if api call invalid, or True if valid.
    """

    # fetch the data
    data = request.json
    print(data)

    # check that all required parameters have been passed
    required_params = ("sender", "message")
    if not all(k in data for k in required_params):
        return return_error(code=400, message=f"Missing one of the required parameters: {required_params}")

    sender_id = data['sender']
    mssg = jsonpickle.decode(data['message'])

    # add the received_messages to the api global variable
    if sender_id not in received_messages:
        received_messages[sender_id] = []
    received_messages[sender_id].append(mssg)

    return jsonify(True)

#########################################################################
# MATRX control api calls
#########################################################################


@app.route('/pause', methods=['GET', 'POST'])
def pause_MATRX():
    """ Pause the MATRX simulation

    Returns
        True if paused, False if already paused
    -------
    """
    global matrx_paused
    if not matrx_paused:
        matrx_paused = True
        return jsonify(True)
    else:
        return jsonify(False)


@app.route('/start', methods=['GET', 'POST'])
def start_MATRX():
    """ Starts / unpauses the MATRX simulation

    Returns
        True if it has been started, False if it is already running
    -------

    """
    global matrx_paused
    if matrx_paused:
        matrx_paused = False
        return jsonify(True)
    else:
        return jsonify(False)


@app.route('/stop', methods=['GET', 'POST'])
def stop_MATRX():
    """ Stops MATRX scenario

    Returns
        True
    -------
    """
    global matrx_done
    matrx_done = True
    return jsonify(True)


@app.route('/change_tick_duration/<tick_dur>', methods=['GET', 'POST'])
def change_MATRX_speed(tick_dur):
    """ Change the tick duration / simulation speed of MATRX

    Parameters
    ----------
    tick_dur
        The duration of 1 tick

    Returns
    -------
        True if successfully changed tick speed (400 error if tick_duration not valid)

    """
    # check if the passed value is a float / int, and return an error if not
    try:
        float(tick_dur)
    except:
        return abort(400, description=f'Tick duration has to be an float, but is of type {type(tick_dur)}')

    # save the new tick duration
    global tick_duration
    tick_duration = float(tick_dur)
    return jsonify(True)


@app.route('/shutdown_API', methods=['GET', 'POST'])
def shutdown():
    """ Shuts down the api by stopping the Flask thread

    Returns
        True
    -------
    """
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Unable to shutdown api server. Not running with the Werkzeug Server')
    func()
    print("api server shutting down...")
    return jsonify(True)


#########################################################################
# Errors
#########################################################################

@app.errorhandler(400)
def bad_request(e):
    print("Throwing error", e)
    return jsonify(error=str(e)), 400


def return_error(code, message):
    """ A helper function that returns a specified error code and message """
    if debug:
        print(f"api request not valid: code {code}. Message: {message}.")
    return abort(code, description=message)


#########################################################################
# api helper methods
#########################################################################


def clean_input_ids(ids):
    """ Clean a received api variable ids to valid Python code

    Parameters
    ----------
    ids
        Can be a string (1 agent id), string encoded list (containing agent ids), list with agent ids, or None

    Returns
    -------
        None or list with string agent IDs

    """
    if ids is None:
        return None

    try:
        ids = eval(ids)
    except:
        pass

    # if it is a list
    if isinstance(ids, list):
        return ids

    if isinstance(ids, str):
        return [ids]


def check_messages_API_request(tick=None, agent_id=None):
    """ Checks if the variables of the api request are valid, and if the requested information exists

    Parameters
    ----------
    tick

    agent_id

    Returns
    -------

    """
    if gw_message_manager is None:
        return False, {'error_code': 400,
                       'error_message': f'MATRX hasn\'t started yet.'}

    tick = current_tick if tick is None else tick
    # check user input, such as tick
    check_passed, error_message = check_input(tick)
    if not check_passed:
        return False, error_message

    return True, None


def check_states_API_request(tick=None, ids=None, ids_required=False):
    """ Checks if the variables of the api request are valid, and if the requested information exists

    Parameters
    ----------
    tick
        MATRX tick
    ids
        string with 1 agent ID, or list of agent IDS
    ids_required
        Whether IDS are required

    Returns
    -------
        Success (Boolean indicating whether it is valid or not), Error (if any, prodiving the type and a message)
        See for the error codes:
        https://www.ibm.com/support/knowledgecenter/SS42VS_7.3.0/com.ibm.qradar.doc/c_rest_api_errors.html


    """
    if gw_message_manager is None:
        return False, {'error_code': 400,
                       'error_message': f'MATRX hasn\'t started yet.'}

    # check user input, such as tick and agent id
    check_passed, error_message = check_input(tick)
    if not check_passed:
        return False, error_message

    # Don't throw an error if MATRX is paused the first tick, and thus still has no states
    # if current_tick is 0 and matrx_paused:
    #     return True, None

    # if this api call requires ids, check this variable on validity as well
    if ids_required:

        # check if ids variable is of a valid type
        try:
            ids = eval(ids)
        except:
            pass
            # return False, {'error_code': 400, 'error_message': f'Provided IDs are not of valid format. Provides IDs
            # is of ' f'type {type(ids)} but should be either of type string for ' f'requesting states of 1 agent (
            # e.g. "god"), or a list of ' f'IDs(string) for requesting states of multiple agents'}

        # check if the api was reset during this time
        if len(states) is 0:
            return False, {'error_code': 400,
                           'error_message': f'api is reconnecting to a new world'}

        # check if the provided ids exist for all requested ticks
        ids = [ids] if isinstance(ids, str) else ids
        for t in range(tick, current_tick + 1):
            for id in ids:
                if id not in states[t]:
                    return False, {'error_code': 400,
                                   'error_message': f'Trying to fetch the state for agent with ID "{id}" for tick {t}, '
                                                    f'but no data on that agent exists for that tick. Is the agent ID '
                                                    f'correct?'}

    # all checks cleared, so return with success and no error message
    return True, None


def check_input(tick=None, ids=None):
    # check if tick is a valid format
    if tick is not None:
        try:
            tick = int(tick)
        except:
            return False, {'error_code': 400,
                           'error_message': f'Tick has to be an integer, but is of type {type(tick)}'}

        # check if the tick has actually occurred
        if not tick in range(0, current_tick + 1):
            return False, {'error_code': 400,
                           'error_message': f'Indicated tick does not exist, has to be in range 0 - {current_tick}, '
                                            f'but is {tick}'}

    return True, None


def __fetch_states(tick, ids=None):
    """ This private function fetches, filters and orders the states as specified by the tick and agent ids.

    Parameters
    ----------
    tick
        Tick from which onwards to return the states. Thus will return a list of length [tick:current_tick]
    ids
        Id(s) from agents/god for which to return the states. Either a single agent ID or a list of agent IDs.
        God view = "god"

    Returns
    -------
        Returns a list of length [tick:current_tick]. For each tick, a dictionary contains the states for each agent as
        specified in `agent_ids`, indexed by their agent ID.

    """
    tick = int(tick)

    # return all states
    if ids is None:
        return states[tick:]

    # convert ints to lists so we can use 1 uniform approach
    try:
        ids = eval(ids)
    except:
        ids = [ids]

    # create a list containing the states from tick to current_tick containing the states of all desired agents/god
    filtered_states = []
    for t in range(tick, current_tick + 1):
        states_this_tick = {}

        # add each agent's state for this tick
        for agent_id in ids:
            states_this_tick[agent_id] = states[t][agent_id]

        # save the states of all filtered agents for this tick
        filtered_states.append(states_this_tick)

    return filtered_states


def __filter_dict(state_dict, props, filters):
    """ Filters a state dictionary to only a dict that contains props for all
    objects that adhere to the filters. A filter is a combination of a
    property and value."""

    def find(obj_dict_pair):
        # Get the object properties
        obj_dict = obj_dict_pair[1]

        # Check if all the desirable properties are in the object dict
        if not all([p in obj_dict.keys() for p in props]):
            return None

        # Check if any filter property is present, if so check its value
        def filter_applies(filter_):
            filter_prop = filter_[0]
            filter_val = filter_[1]
            if filter_prop in obj_dict.keys():
                return filter_val == obj_dict[filter_prop] \
                       or filter_val in obj_dict[filter_prop]
            else:
                return False  # if filter is not present, we return False

        # If filters are given, go over each filter to see if it applies
        if filters is not None:
            filter_results = map(filter_applies, filters.items())

            # Check if all filters are either not applicable or return true
            applies = all(filter_results)
            if applies is False:  # if it does not adhere to the filters
                return None

        # Filter the dict to only the required properties
        new_dict = {p: obj_dict[p] for p in props}

        # Return the new tuple
        return obj_dict_pair[0], new_dict

    # Map our find method to all state objects
    filtered_objects = map(find, state_dict.items())

    # Extract all state objects that have the required properties and adhere to
    # the filters
    objects = [obj_dict_pair for obj_dict_pair in filtered_objects
               if obj_dict_pair is not None]

    # Transform back to dict
    objects = {obj_id: obj_dict for obj_id, obj_dict in objects}

    # Return
    return objects


def create_error_response(code, message):
    """ Creates an error code with a custom message """
    response = jsonify({'message': message})
    response.status_code = code
    return response


def __reorder_state(state):
    """ This private function makes the MATRX state ready for sending as a JSON object

    Parameters
    ----------
    state
         The world state, a dictionary with object IDs as keys

    Returns
    -------
        The world state, JSON serializable
    """
    new_state = copy.copy(state)

    # loop through all objects in the state
    for objID, obj in state.items():

        if objID is not "World":
            # make the sense capability JSON serializable
            if "sense_capability" in obj:
                new_state[objID]["sense_capability"] = str(obj["sense_capability"])

    return new_state


def add_state(agent_id, state, agent_inheritence_chain, world_settings):
    """ Saves the state of an agent for use via the api

    Parameters
    ----------
    agent_id
         ID of the agent of who the state is
    state
        state as filtered by the agent
    agent_inheritence_chain
         inheritance_chain of classes, can be used to figure out type of agent
    world_settings
        This object contains all information on the MATRX world, such as tick and grid_size. Some agents might filter
        these out of their state, as such it is sent along seperatly to make sure the world settings, required by the
        visualization, are passed along.
    -------
    """
    # save the new general info on the MATRX World (once)
    global next_tick_info
    if next_tick_info == {}:
        next_tick_info = world_settings

    # Make sure the world settings are in the state, as these are used by the visualization
    if 'World' not in state:
        state['World'] = world_settings

    # state['World']['matrx_paused'] = matrx_paused

    # reorder and save the new state along with some meta information
    temp_state[agent_id] = {'state': __reorder_state(state), 'agent_inheritence_chain': agent_inheritence_chain}


def next_tick():
    """ Proceed to the next tick, publicizing data of the new tick via the api (the new states).
    -------
    """
    # save the new general info
    global MATRX_info, next_tick_info
    MATRX_info = copy.copy(next_tick_info)
    next_tick_info = {}
    # print("Next ticK:", MATRX_info);

    # publicize the states of the previous tick
    states.append(copy.copy(temp_state))


def pop_userinput(agent_id):
    """ Pop the user input for an agent from the userinput dictionary and return it

    Parameters
    ----------
    agent_id
        ID of the agent for which to return the userinput

    Returns
    -------
        A list of keys pressed. See this link for the encoding of keys:
        https://developer.mozilla.org/nl/docs/Web/API/KeyboardEvent/key/Key_Values
    """
    global userinput
    return userinput.pop(agent_id, None)


def reset_api():
    """ Reset the MATRX api variables """
    global temp_state, userinput, matrx_paused, matrx_done, states, current_tick, tick_duration, grid_size
    global MATRX_info, next_tick_info, received_messages, current_world_ID
    temp_state = {}
    userinput = {}
    matrx_paused = False
    matrx_done = False
    states = []
    current_tick = 0
    tick_duration = 0.0
    grid_size = [1, 1]
    MATRX_info = {}
    next_tick_info = {}
    received_messages = {}
    current_world_ID = False


def register_world(world_id):
    """ Register a new simulation world

    At the moment simulation of only one world at a time is supported, so this calling this function will discard
    the previous world.

    Parameters
    ----------
    world_id
        The ID of the world
    -------
    """
    global current_world_ID
    current_world_ID = world_id


#########################################################################
# api Flask methods
#########################################################################

def flask_thread():
    """ Starts the Flask server on localhost:3001
    -------
    """
    if not debug:
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    app.run(host='0.0.0.0', port=port, debug=False, use_reloader=False)


def run_api(verbose=False):
    """ Creates a separate Python thread in which the api (Flask) is started
    Returns
        MATRX api Python thread
    -------
    """
    print("Starting background api server")
    global debug
    debug = verbose

    print("Initialized app:", app)
    api_thread = threading.Thread(target=flask_thread)
    api_thread.start()
    return api_thread


if __name__ == "__main__":
    run_api()



import itertools
import math


def get_distance(coord1, coord2):
    """ Get distance between two x,y coordinates """
    dist = [(a - b) ** 2 for a, b in zip(coord1, coord2)]
    dist = math.sqrt(sum(dist))
    return dist


def get_room_locations(room_top_left, room_width, room_height):
    """ Returns the locations within a room, excluding walls.

    This is a helper function for adding objects to a room. It returns a
    list of all (x,y) coordinates that fall within the room excluding the
    walls.

    Parameters
    ----------
    room_top_left: tuple, (x, y)
        The top left coordinates of a room, as used to add that room with
        methods such as `add_room`.
    room_width: int
        The width of the room.
    room_height: int
        The height of the room.

    Returns
    -------
    list, [(x,y), ...]
        A list of (x, y) coordinates that are encapsulated in the
        rectangle, excluding walls.

    See Also
    --------
    WorldBuilder.add_room

    """
    xs = list(range(room_top_left[0] + 1,
                    room_top_left[0] + room_width - 1))
    ys = list(range(room_top_left[1] + 1,
                    room_top_left[1] + room_height -1))
    locs = list(itertools.product(xs, ys))
    return locs


def __flatten_dict(dict_):
    new_dict = {}
    for k, v in dict_.items():
        if isinstance(v, dict):
            sub_dict_ = {f"{k}_{k2}": v2 for k2, v2 in v.items()}
            new_dict = {**new_dict, **sub_dict_}
        else:
            new_dict[k] = v

    return new_dict


# from matrx.goals.goals import LimitedTimeGoal

######################
# AgentBody defaults #
######################
AGENTBODY_IS_TRAVERSABLE = False
AGENTBODY_IS_MOVABLE = False
AGENTBODY_SPEED_IN_TICKS = 1
AGENTBODY_VIS_SIZE = 1.0
AGENTBODY_VIS_COLOUR = "#92f441"
AGENTBODY_VIS_OPACITY = 1.0
AGENTBODY_VIS_SHAPE = 1
AGENTBODY_VIS_DEPTH = 100
AGENTBODY_VIS_BUSY = False
AGENTBODY_POSSIBLE_ACTIONS = "*"

######################
# EnvObject defaults #
######################
ENVOBJECT_IS_TRAVERSABLE = False
ENVOBJECT_IS_MOVABLE = True
ENVOBJECT_VIS_SIZE = 1.0
ENVOBJECT_VIS_SHAPE = 0
ENVOBJECT_VIS_COLOUR = "#4286f4"
ENVOBJECT_VIS_OPACITY = 1.0
ENVOBJECT_VIS_DEPTH = 80

######################
# GridWorld defaults #
######################
GRIDWORLD_RANDOM_SEED = 1
GRIDWORLD_SIZE = [25, 25]
GRIDWORLD_STEP_DURATION = 0.1
GRIDWORLD_TIME_FOCUS = "step"
GRIDWORLD_RUN_VIS_SERVER = False
GRIDWORLD_RUN_API = False
# GRIDWORLD_SIMULATION_GOAL = LimitedTimeGoal
GRIDWORLD_SIM_GOAL_ARGUMENTS = {
    "max_nr_ticks": -1
}



import datetime
import os.path
import warnings
from collections import OrderedDict
import time
import copy

import gevent

from matrx.actions.object_actions import *
from matrx.logger.logger import GridWorldLogger
from matrx.objects.env_object import EnvObject
from matrx.objects.standard_objects import AreaTile
from matrx.messages.message_manager import MessageManager
from matrx.objects.agent_body import _get_all_classes
from matrx.api import api


class GridWorld:
    """  The Gridworld is the representation of the world and the core of MATRX
    """

    def __init__(self, shape, tick_duration, simulation_goal, rnd_seed=1,
                 visualization_bg_clr="#C2C2C2", visualization_bg_img=None, verbose=False, world_id=0):

        """ Create a GridWorld instance.

        With the constructor you can set a number of general properties and
        from the resulting instance you can call numerous methods to add new
        objects and/or agents.

        Parameters
        ----------
        shape : tuple or list
           Denotes the width and height of the world you create.

        tick_duration : float
           The duration of a single 'tick' or loop in the game-loop of the
           world you create.

        simulation_goal : WorldGoal, int, list or tuple
           The goal or goals of the world, either a single `WorldGoal`, a
           list of such or a positive non-zero integer to denote the maximum
           number of 'ticks' the world(s) has to run.

        rnd_seed : bool (optional, default, 1)
            The master random seed set by the WorldBuilder on which all objects, agents and worlds are
            seeded. Should be a positive non-zero integer.

        visualization_bg_clr : string (optional, "C2C2C2")
           The color of the world when visualized using MATRX' own
           visualisation server. A string representation of hexadecimal color.

        visualization_bg_img : string (optional, None)
           An optional background image of the world when visualized using
           MATRX' own visualisation server. A string of the path to the image
           file. When None, no background image is used.

        verbose : bool (optional, False)
           Whether the GridWorld should be verbose and print development logs to the console.

        world_id : int (optional, 0)
           The ID of this world. Every new GridWorld instance should have a unique ID, such that the frontend knows
           when it has to reinitialize the visualization.


        Examples
        --------

        Create a world builder and generate 10 worlds and run them:
        >>> from matrx.world_builder import WorldBuilder
        >>> builder = WorldBuilder(shape=(10, 10))
        >>> for gridworld in builder.worlds():
        >>>     gridworld.run()

        Every `gridworld` received from builder.worlds() is a GridWorld instance.

        """

        self.__tick_duration = tick_duration  # How long each tick should take (process sleeps until this time passed)
        self.__shape = shape  # The width and height of the GridWorld
        self.__visualization_bg_clr = visualization_bg_clr  # The background color of the visualisation
        self.__visualization_bg_img = visualization_bg_img  # The background image of the visualisation
        self.__verbose = verbose  # Set whether we should print anything or not
        self.world_id = world_id  # ID of this simulation world

        self.__teams = {}  # dictionary with team names (keys), and agents in those teams (values)
        self.__registered_agents = OrderedDict()  # The dictionary of all existing agents in the GridWorld
        self.__environment_objects = OrderedDict()  # The dictionary of all existing objects in the GridWorld

        # The simulation goal, the simulation ends when this/these are reached.
        # Copy and reset all simulation goals, this to make sure that this world has its own goals independent of any
        # other world that potentially received the same simulation goal instance (or list of them)
        if isinstance(simulation_goal, (tuple, list)):
            self.__simulation_goal = [sg.reset() for sg in simulation_goal]
        else:
            self.__simulation_goal = simulation_goal.reset()

        # Get all actions within all currently imported files
        self.__all_actions = _get_all_classes(Action, omit_super_class=True)

        # Initialise an empty grid, a simple 2D array with ID's
        self.__grid = np.array([[None for _ in range(shape[0])] for _ in range(shape[1])])

        self.__api_info = None  # Dict containing info about the API instance
        self.__run_matrx_api = False  # Bool if API is running
        self.__loggers = []  # a list of GridWorldLogger use to log the data
        self.__is_done = False  # Whether the simulation is done (goal(s) reached)
        self.__rnd_seed = rnd_seed  # The random seed of this GridWorld
        self.__rnd_gen = np.random.RandomState(seed=self.__rnd_seed)  # The random state of this GridWorld
        self.__curr_tick_duration = 0.  # Duration of the current tick
        self.__current_nr_ticks = 0  # The number of tick this GridWorld has ran already
        self.__is_initialized = False  # Whether this GridWorld is already initialized
        self.__message_buffer = {}  # dictionary of messages that need to be send to agents, with receiver ids as keys
        self.message_manager = MessageManager()  # keeps track of all messages and makes them available to the api

    def initialize(self, api_info):
        """ Initializes the gridworld instance and any connected visualizations via the API, then pauses the GridWorld.

        By default the GridWorld starts paused. To make it possible for the user to see the first tick of the
        GridWorld, before pressing play and starting the simulation, this function initializes the gridworld to
        tick 0, after which it updates any visualizations via the API, followed by pausing the GridWorld.

        Parameters
        ----------
        api_info : dictionary
           A dictionary which contians information on the API. At least containing the `api_thread` key with the
           API thread, and the `run_matrx_api` key which contains whether the GridWorld should start paused or not.

        Examples
        --------

        Create a world builder and generate 10 worlds and run them:
        >>> from matrx.world_builder import WorldBuilder
        >>> builder = WorldBuilder(shape=(10, 10))
        >>> for gridworld in builder.worlds():
        >>>     gridworld.run()

        Every `gridworld` received from builder.worlds() is a GridWorld instance.

        """

        # Only initialize when we did not already do so
        if not self.__is_initialized:
            # We update the grid, which fills everything with added objects and agents
            self.__update_grid()

            for agent_body in self.__registered_agents.values():
                agent_body.brain_initialize_func()

            # set the api variables
            self.__api_info = api_info
            self.__run_matrx_api = self.__api_info['run_matrx_api']
            if self.__run_matrx_api:
                # initialize this world in the api
                api.reset_api()
                api.tick_duration = self.__tick_duration
                api.register_world(self.world_id)
                api.current_tick = self.__current_nr_ticks
                api.grid_size = self.shape
                # point the api towards our message manager, for making messages available via the api
                api.gw_message_manager = self.message_manager
                api.gw = self
                api.teams = self.__teams

                # init api with world info
                api.MATRX_info = {
                    "nr_ticks": self.__current_nr_ticks,
                    "curr_tick_timestamp": int(round(time.time() * 1000)),
                    "grid_shape": self.__shape,
                    "tick_duration": self.tick_duration,
                    "world_ID": self.world_id,
                    "vis_settings": {
                        "vis_bg_clr": self.__visualization_bg_clr,
                        "vis_bg_img": self.__visualization_bg_img
                    }
                }
                # start paused
                api.matrx_paused = True

                # fetch the initial state of every agent to display
                self._fetch_initial_states()

            # Set initialisation boolean
            self.__is_initialized = True

            if self.__verbose:
                print(f"@{os.path.basename(__file__)}: Initialized the GridWorld.")

    def run(self, api_info):
        """ Runs the gridworld instance until stopped via the visualization or the goal has been achieved.

        The gridworld by default starts paused.

        Parameters
        ----------
        api_info : dictionary
          A dictionary which contians information on the API. At least containing the `api_thread` key with the
          API thread, and the `run_matrx_api` key which contains whether the GridWorld should start paused or not.

        Examples
        --------

        Create a world builder and generate 10 worlds and run them:
        >>> from matrx.world_builder import WorldBuilder
        >>> builder = WorldBuilder(shape=(10, 10))
        >>> for gridworld in builder.worlds():
        >>>     gridworld.run()

        """

        # initialize the gridworld
        self.initialize(api_info)

        if self.__verbose:
            print(f"@{os.path.basename(__file__)}: Starting game loop...")
        is_done = False
        while not is_done:

            if self.__run_matrx_api and api.matrx_paused:
                print("MATRX paused through api")
                gevent.sleep(1)
            else:
                is_done, tick_duration = self.__step()

            if self.__run_matrx_api and api.matrx_done:
                print("Scenario stopped through api")
                break

    def get_env_object(self, requested_id, obj_type=None):
        """ Fetch an object or agent from the GridWorld using its ID, optionally checking for its object type.

        Parameters
        ----------
        requested_id : dictionary
          The ID of the object or agent to fetch.

        obj_type : Class (optional, default, None)
          The object class of which the  object-to-fetch should be. Can be used as a filter: if the object with id
          `requested_id` is not of this class, it is not returned.

        Returns
        -------
        obj : Object or Agent
            Returns the object or agent with ID `requested_id` of tpe `obj_type` (if not None). If no object can be
            found that adheres to these filters, None is returned.

        Examples
        --------

        In an action, world goal, or somewhere else with access to the Gridworld, the function can be used as below.
        In this example the custom action removes a specific agent from the world when executed.

        >>> from matrx.agents import HumanAgent
        >>> class YourCustomAction(Action):
        >>>     def __init__(...):
        >>>         ...
        >>>
        >>>     def mutate(self, grid_world, agent_id, **kwargs):
        >>>         agent = grid_world.get_env_object("HumanAgent_23", HumanAgent)
        >>>         del agent

        Alternatively, if you know the object ID and type (object or agent) for certain, it can also be fetched
        as so:

        >>> reg_ag = grid_world.registered_agents[agent_id] # fetch agent by ID
        >>> env_obj = grid_world.environment_objects[object_id]  # fetch object by ID

        """

        obj = None

        # try to fetch object from agents
        if requested_id in self.__registered_agents.keys():
            if obj_type is not None:
                if isinstance(self.__registered_agents[requested_id], obj_type):
                    obj = self.__registered_agents[requested_id]
            else:
                obj = self.__registered_agents[requested_id]

        # try to fetch object from env objects
        if requested_id in self.__environment_objects.keys():
            if obj_type is not None:
                if isinstance(self.__environment_objects[requested_id], obj_type):
                    obj = self.__environment_objects[requested_id]
            else:
                obj = self.__environment_objects[requested_id]

        return obj

    def get_objects_in_range(self, agent_loc, object_type, sense_range):
        """ Get all objects of a specific obj type (normal objects or agent) within a certain range around an
        agent's location.

        Parameters
        ----------
        agent_loc : dictionary
          The location from which to search in a radius around for objects. Can be any [x,y] location within the
          GridWorld shape dimensions.

        obj_type : Class (optional, default, None)
          The object class of which the objects-to-find should be.

        sense_range : int
          The radius around the agent within which to look for objects.

        Returns
        -------
        obj : OrderedDict
            Returns an ordereddict with the objects and agents of tpe `object_type` within `sense_range` blocks of the
            location passed via `agent_loc`. If not objects were found, the ordered dict is empty.

        Examples
        --------

        In an action, world goal, or somewhere else with access to the Gridworld, the function can be used as
        below.
        In this example all objects of all types (denoted with the "*") within 5 tiles of the coordinate [3,3] are
        returned.

        >>> objects_in_range = grid_world.get_objects_in_range([3,3], object_type="*", sense_range=5)

        """

        env_objs = OrderedDict()
        # loop through all environment objects
        for obj_id, env_obj in self.__environment_objects.items():
            # get the distance from the agent location to the object
            coordinates = env_obj.location
            distance = get_distance(coordinates, agent_loc)

            # check if the env object is of the specified type, and within range
            if (object_type is None or object_type == "*" or isinstance(env_obj, object_type)) and \
                    distance <= sense_range:
                env_objs[obj_id] = env_obj

        # agents are also environment objects, but stored separably. Also check them.
        for agent_id, agent_obj in self.__registered_agents.items():
            coordinates = agent_obj.location
            distance = get_distance(coordinates, agent_loc)

            # check if the env object is of the specified type, adn within range
            if (object_type is None or object_type == "*" or isinstance(agent_obj, object_type)) and \
                    distance <= sense_range:
                env_objs[agent_id] = agent_obj
        return env_objs

    def remove_from_grid(self, object_id, remove_from_carrier=True):
        """ Remove an object from the grid.

        Parameters
        ----------
        object_id : int
          ID of the object to remove

        remove_from_carrier : Bool (optional, default, True)
          Whether to also remove from agents that are currently carrying the object.

        Returns
        -------
        obj : Bool
            Whether the object was sucessfully removed from the grid or not.

        Examples
        --------

        In an action, world goal, or somewhere else with access to the Gridworld, the function can be used as
        below.
        In this example the Agent3 is removed from the Grid.
        >>> succeeded = grid_world.remove_from_grid(object_id="Agent3", remove_from_carrier=False)

        In this example the object Block5 is removed from the Grid, and any Agents that were carrying it.
        >>> succeeded = grid_world.remove_from_grid(object_id="Block5", remove_from_carrier=True)

        """


        # Remove object first from grid
        grid_obj = self.get_env_object(object_id)  # get the object
        loc = grid_obj.location  # its location

        self.__grid[loc[1], loc[0]].remove(grid_obj.obj_id)  # remove the object id from the list at that location
        if len(self.__grid[loc[1], loc[0]]) == 0:  # if the list is empty, just add None there
            self.__grid[loc[1], loc[0]] = None

        # Remove object from the list of registered agents or environmental objects
        # Check if it is an agent
        if object_id in self.__registered_agents.keys():
            # Check if the agent was carrying something, if so remove property from carried item
            for obj_id in self.__registered_agents[object_id].is_carrying:
                self.__environment_objects[obj_id].carried_by.remove(object_id)

            # Remove agent
            success = self.__registered_agents.pop(object_id,
                                                   default=False)  # if it exists, we get it otherwise False

        # Else, check if it is an object
        elif object_id in self.__environment_objects.keys():
            # remove from any agents carrying this object if asked for
            if remove_from_carrier:
                # If the object was carried, remove this from the agent properties as well
                for agent_id in self.__environment_objects[object_id].carried_by:
                    obj = self.__environment_objects[object_id]
                    self.__registered_agents[agent_id].is_carrying.remove(obj)

            # Remove object
            success = self.__environment_objects.pop(object_id,
                                                     default=False)  # if it exists, we get it otherwise False
        else:
            success = False  # Object type not specified

        if success is not False:  # if succes is not false, we successfully removed the object from the grid
            success = True

        if self.__verbose:
            if success:
                print(f"@{os.path.basename(__file__)}: Succeeded in removing object with ID {object_id}")
            else:
                print(f"@{os.path.basename(__file__)}: Failed to remove object with ID {object_id}.")

        return success

    def _fetch_initial_states(self):
        """ MATRX starts paused by default, to prime the api and any connected GUI's, we fetch the first state
        from all agents to send which can be shown while waiting for the experiment leader to press play.
        """

        for agent_id, agent_obj in self.__registered_agents.items():
            # given the agent's capabilities, get everything the agent can perceive
            state = self.__get_agent_state(agent_obj)

            # filter other things from the agent state
            filtered_agent_state = agent_obj.filter_observations(state)

            # save the current agent's state for the api
            api.add_state(agent_id=agent_id, state=filtered_agent_state,
                          agent_inheritence_chain=agent_obj.class_inheritance,
                          world_settings=api.MATRX_info)

        # add god state
        api.add_state(agent_id="god", state=self.__get_complete_state(), agent_inheritence_chain="god",
                      world_settings=api.MATRX_info)

        # initialize the message manager
        self.message_manager.agents = self.__registered_agents.keys()
        self.message_manager.teams = self.__teams

        # make the information of this tick available via the api, after all
        # agents have been updated
        api.next_tick()

    def _register_agent(self, agent, agent_body: AgentBody):
        """ Register human agents and agents to the gridworld environment """

        # Random seed for agent between 1 and 10000000, might need to be adjusted still
        agent_seed = self.__rnd_gen.randint(1, 1000000)

        # check if the agent can be succesfully placed at that location
        self.__validate_obj_placement(agent_body)

        # Add agent to registered agents
        self.__registered_agents[agent_body.obj_id] = agent_body

        if self.__verbose:
            print(f"@{os.path.basename(__file__)}: Created agent with id {agent_body.obj_id}.")

        # Get all properties from the agent avatar
        avatar_props = agent_body.properties

        if agent_body.is_human_agent is False:
            agent._factory_initialise(agent_name=agent_body.obj_name,
                                      agent_id=agent_body.obj_id,
                                      action_set=agent_body.action_set,
                                      sense_capability=agent_body.sense_capability,
                                      agent_properties=avatar_props,
                                      customizable_properties=agent_body.customizable_properties,
                                      callback_is_action_possible=self.__check_action_is_possible,
                                      rnd_seed=agent_seed)
        else:  # if the agent is a human agent, we also assign its user input action map
            agent._factory_initialise(agent_name=agent_body.obj_name,
                                      agent_id=agent_body.obj_id,
                                      action_set=agent_body.action_set,
                                      sense_capability=agent_body.sense_capability,
                                      agent_properties=avatar_props,
                                      customizable_properties=agent_body.customizable_properties,
                                      callback_is_action_possible=self.__check_action_is_possible,
                                      rnd_seed=agent_seed,
                                      key_action_map=agent_body.properties["key_action_map"])

        return agent_body.obj_id

    def _register_env_object(self, env_object: EnvObject):
        """ this function adds the objects """

        # check if the object can be succesfully placed at that location
        self.__validate_obj_placement(env_object)

        # Assign id to environment sparse dictionary grid
        self.__environment_objects[env_object.obj_id] = env_object

        if self.__verbose:
            print(f"@{__file__}: Created an environment object with id {env_object.obj_id}.")

        return env_object.obj_id

    def _register_teams(self):
        """ Register all teams and who is in those teams.
        An agent is always in a team, if not set by the user, a team is created with name 'agent_id' with only that
        agent in it.
        """
        # loop through all agents
        for agent_id, agent_body in self.registered_agents.items():
            # find their team name
            team = agent_body.properties['team']

            # register the team (if not already done) and the agent in it
            if team not in self.__teams:
                self.__teams[team] = []
            self.__teams[team].append(agent_id)

    def _register_logger(self, logger: GridWorldLogger):
        if self.__loggers is None:
            self.__loggers = [logger]
        else:
            self.__loggers.append(logger)

    def __add_to_grid(self, grid_obj):

        if isinstance(grid_obj, EnvObject):
            loc = grid_obj.location
            if self.__grid[loc[1], loc[0]] is not None:
                self.__grid[loc[1], loc[0]].append(grid_obj.obj_id)
            else:
                self.__grid[loc[1], loc[0]] = [grid_obj.obj_id]
        else:
            raise BaseException(f"Object is not of type {str(type(EnvObject))} but of {str(type(grid_obj))} when adding"
                                f" to grid in GridWorld.")

    def __validate_obj_placement(self, env_object):
        """
        Checks whether an object can be successfully placed on the grid
        """
        obj_loc = env_object.location

        # get the objects at the target object location
        objs_at_loc = self.get_objects_in_range(obj_loc, "*", 0)

        # filter out areaTiles, which don't count
        for key in list(objs_at_loc.keys()):
            if AreaTile.__name__ in objs_at_loc[key].class_inheritance:
                objs_at_loc.pop(key)

        # check how many of these objects are intraversable
        intraversable_objs = []
        for obj in objs_at_loc:
            if not objs_at_loc[obj].is_traversable:
                intraversable_objs.append(objs_at_loc[obj].obj_id)

        # two intraversable objects can't be at the same location
        if not env_object.is_traversable and len(intraversable_objs) > 0:
            raise Exception(f"Invalid placement. Could not place object {env_object.obj_id} in grid, location already "
                            f"occupied by intraversable object {intraversable_objs} at location {obj_loc}")

    def __step(self):

        # Set tick start of current tick
        start_time_current_tick = datetime.datetime.now()

        # Check if we are done based on our global goal assessment function
        self.__is_done, goal_status = self.__check_simulation_goal()

        # Log the data if we have any loggers
        for logger in self.__loggers:
            agent_data_dict = {}
            for agent_id, agent_body in self.__registered_agents.items():
                agent_data_dict[agent_id] = agent_body.get_log_data()

            logger._grid_world_log(grid_world=self, agent_data=agent_data_dict,
                                   last_tick=self.__is_done, goal_status=goal_status)

        # If this grid_world is done, we return immediately
        if self.__is_done:
            return self.__is_done, 0.

        # initialize a temporary dictionary in which all states of this tick
        # will be saved. After all agents have been updated, the new tick info
        # will be made accessible via the api.
        if self.__run_matrx_api:
            api.temp_state = {}

            # if this is the first tick, clear the placeholder states
            if self.__current_nr_ticks == 0:
                api.MATRX_info = {}
                api.next_tick_info = {}

        # Go over all agents, detect what each can detect, figure out what actions are possible and send these to
        # that agent. Then receive the action back and store the action in a buffer.
        # Also, update the local copy of the agent properties, and save the agent's state for the GUI.
        # Then go to the next agent.
        # This blocks until a response from the agent is received (hence a tick can take longer than self.tick_
        # duration!!)
        action_buffer = OrderedDict()
        for agent_id, agent_obj in self.__registered_agents.items():

            state = self.__get_agent_state(agent_obj)

            # check if this agent is busy performing an action , if so then also check if it as its last tick of waiting
            # because then we want to do that action. If not busy, call its get_action function.
            if agent_obj._check_agent_busy(curr_tick=self.__current_nr_ticks):

                # only do the filter observation method to be able to update the agent's state to the api
                filtered_agent_state = agent_obj.filter_observations(state)

                # save the current agent's state for the api
                if self.__run_matrx_api:
                    api.add_state(agent_id=agent_id, state=filtered_agent_state,
                                  agent_inheritence_chain=agent_obj.class_inheritance,
                                  world_settings=self.__get_complete_state()['World'])

            else:  # agent is not busy

                # Any received data from the api for this HumanAgent is send along to the get_action function
                if agent_obj.is_human_agent:
                    usrinp = None
                    if self.__run_matrx_api and agent_id in api.userinput:
                        usrinp = api.pop_userinput(agent_id)

                    filtered_agent_state, agent_properties, action_class_name, action_kwargs = \
                        agent_obj.get_action_func(state=state, agent_properties=agent_obj.properties, agent_id=agent_id,
                                                  user_input=usrinp)
                else:  # not a HumanAgent

                    # perform the agent's get_action method (goes through filter_observations and decide_on_action)
                    filtered_agent_state, agent_properties, action_class_name, action_kwargs = agent_obj.get_action_func(
                        state=state, agent_properties=agent_obj.properties, agent_id=agent_id)

                # the Agent (in the OODA loop) might have updated its properties, process these changes in the Avatar
                # Agent
                agent_obj._set_agent_changed_properties(agent_properties)

                # Set the agent to busy, we do this only here and not when the agent was already busy to prevent the
                # agent to perform an action with a duration indefinitely (and since all actions have a duration, that
                # would be killing...)
                self.__set_agent_busy(action_name=action_class_name, action_kwargs=action_kwargs, agent_id=agent_id)

                # Get all agents we have, as we need these to process all messages that are send to all agents
                all_agent_ids = self.__registered_agents.keys()

                # Obtain all communication messages if the agent has something to say to others (only comes here when
                # the agent is NOT busy)
                agent_messages = agent_obj.get_messages_func(all_agent_ids)

                # add any messages received from the api sent by this agent
                if self.__run_matrx_api:
                    if agent_id in api.received_messages:
                        agent_messages += copy.copy(api.received_messages[agent_id])

                        # clear the messages for the next tick
                        del api.received_messages[agent_id]

                # preprocess all messages of the current tick of this agent
                self.message_manager.preprocess_messages(self.__current_nr_ticks, agent_messages,
                                                         all_agent_ids, self.__teams)

            # save the current agent's state for the api
            if self.__run_matrx_api:
                api.add_state(agent_id=agent_id, state=filtered_agent_state,
                              agent_inheritence_chain=agent_obj.class_inheritance,
                              world_settings=self.__get_complete_state()['World'])

            # if this agent is at its last tick of waiting on its action duration, we want to actually perform the
            # action
            if agent_obj._at_last_action_duration_tick(curr_tick=self.__current_nr_ticks):
                # Get the action and arguments
                action_class_name, action_kwargs = agent_obj._get_duration_action()
                # store the action in the buffer
                action_buffer[agent_id] = (action_class_name, action_kwargs)

        # put all messages of the current tick in the message buffer
        if self.__current_nr_ticks in self.message_manager.preprocessed_messages:
            for mssg in self.message_manager.preprocessed_messages[self.__current_nr_ticks]:
                if mssg.to_id not in self.__message_buffer.keys():  # first message for this receiver
                    self.__message_buffer[mssg.to_id] = [mssg]
                else:
                    self.__message_buffer[mssg.to_id].append(mssg)

        # save the god view state
        if self.__run_matrx_api:
            api.add_state(agent_id="god", state=self.__get_complete_state(), agent_inheritence_chain="god",
                          world_settings=self.__get_complete_state()['World'])

            # make the information of this tick available via the api, after all
            # agents have been updated
            api.next_tick()
            api.current_tick = self.__current_nr_ticks
            self.__tick_duration = api.tick_duration
            api.grid_size = self.shape

        # Perform the actions in the order of the action_buffer (which is filled in order of registered agents
        for agent_id, action in action_buffer.items():
            # Get the action class name
            action_class_name = action[0]
            # Get optional kwargs
            action_kwargs = action[1]

            if action_kwargs is None:  # If kwargs is none, make an empty dict out of it
                action_kwargs = {}

            # Actually perform the action (if possible), also sets the result in the agent's brain
            self.__perform_action(agent_id, action_class_name, action_kwargs)

            # Update the grid
            self.__update_grid()

        # Send all messages between agents
        for receiver_id, messages in self.__message_buffer.items():
            # check if the receiver exists
            if receiver_id in self.__registered_agents.keys():
                # Call the callback method that sets the messages
                self.__registered_agents[receiver_id].set_messages_func(messages)

        self.__message_buffer = {}

        # Perform the update method of all objects
        for env_obj in self.__environment_objects.values():
            env_obj.update(self)

        # Increment the number of tick we performed
        self.__current_nr_ticks += 1

        # Check how much time the tick lasted already
        tick_end_time = datetime.datetime.now()
        tick_duration = tick_end_time - start_time_current_tick
        self.sleep_duration = self.__tick_duration - tick_duration.total_seconds()

        # Sleep for the remaining time of self.__tick_duration
        self.__sleep()

        # Compute the total time of our tick (including potential sleep)
        tick_end_time = datetime.datetime.now()
        tick_duration = tick_end_time - start_time_current_tick
        self.__curr_tick_duration = tick_duration.total_seconds()

        if self.__verbose:
            print(f"@{os.path.basename(__file__)}: Tick {self.__current_nr_ticks} took {tick_duration.total_seconds()} "
                  f"seconds.")

        return self.__is_done, self.__curr_tick_duration

    def __check_simulation_goal(self):

        goal_status = {}
        if self.__simulation_goal is not None:
            if isinstance(self.__simulation_goal, (list, tuple)):  # edited this check to include tuples
                for sim_goal in self.__simulation_goal:
                    is_done = sim_goal.goal_reached(self)
                    goal_status[sim_goal] = is_done
            else:
                is_done = self.__simulation_goal.goal_reached(self)
                goal_status[self.__simulation_goal] = is_done

        is_done = np.array(list(goal_status.values())).all()
        return is_done, goal_status

    def __sleep(self):
        """
        Sleeps the current python process for the amount of time that is left after self.curr_tick_duration up to
        in self.__tick_duration
        :return:
        """
        if self.sleep_duration > 0:
            gevent.sleep(self.sleep_duration)
        else:
            self.__warn(
                f"The average tick took longer than the set tick duration of {self.__tick_duration}. "
                f"Program is to heavy to run real time")

    def __update_grid(self):
        self.__grid = np.array([[None for _ in range(self.__shape[0])] for _ in range(self.__shape[1])])
        for obj_id, obj in self.__environment_objects.items():
            self.__add_to_grid(obj)
        for agent_id, agent in self.__registered_agents.items():
            self.__add_to_grid(agent)

    # get all objects and agents on the grid
    def __get_complete_state(self):
        """
        Compile all objects and agents on the grid in one state dictionary
        :return: state with all objects and agents on the grid
        """

        # create a state with all objects and agents
        state = {}
        for obj_id, obj in self.__environment_objects.items():
            state[obj.obj_id] = obj.properties
        for agent_id, agent in self.__registered_agents.items():
            state[agent.obj_id] = agent.properties

        # Append generic properties (e.g. number of ticks, size of grid, etc.}
        state["World"] = {
            "nr_ticks": self.__current_nr_ticks,
            "curr_tick_timestamp": int(round(time.time() * 1000)),
            "grid_shape": self.__shape,
            "tick_duration": self.tick_duration,
            "world_ID": self.world_id,
            "vis_settings": {
                "vis_bg_clr": self.__visualization_bg_clr,
                "vis_bg_img": self.__visualization_bg_img
            }
        }

        return state

    def __get_agent_state(self, agent_obj: AgentBody):
        agent_loc = agent_obj.location
        sense_capabilities = agent_obj.sense_capability.get_capabilities()

        # Check if sense capabilities contains the "*" wildcard. If so we store
        # those objects separately, in case there are also other types specified
        # whose range precedes that of the "*" wildcard. For example, when an
        # agent can perceive other agents within a certain range and also has
        # the "*" wildcard to denote all OTHER objects. In that case we should
        # ignore all the AgentBody objects in the "*"-range. and then add all
        # the AgentBody objects we can see with the "AgentBody"-range.
        wildcard_objs = {}
        objs_in_range = OrderedDict()
        if "*" in sense_capabilities.keys():
            wildcard_objs = self.get_objects_in_range(agent_loc, "*", sense_capabilities["*"])
            sense_capabilities.pop("*")  # we did the wildcard already

        # Check which objects can be sensed with the agents' capabilities, from
        # its current position.
        for obj_type, sense_range in sense_capabilities.items():
            env_objs = self.get_objects_in_range(agent_loc, obj_type, sense_range)
            objs_in_range.update(env_objs)

        # Merge the wildcard objects with those of the (potential) other objects,
        # where we ignore all wildcard objects that are of a specific given type
        # with its own range.
        for wildcard_obj_id, wildcard_obj in wildcard_objs.items():
            if type(wildcard_obj) not in sense_capabilities.keys():
                objs_in_range[wildcard_obj_id] = wildcard_obj

        state = {}
        # Save all properties of the sensed objects in a state dictionary
        for env_obj in objs_in_range:
            state[env_obj] = objs_in_range[env_obj].properties

        # Append generic properties (e.g. number of ticks, fellow team members, etc.}
        team_members = [agent_id for agent_id, other_agent in self.__registered_agents.items()
                        if agent_obj.team == other_agent.team]
        state["World"] = {
            "nr_ticks": self.__current_nr_ticks,
            "curr_tick_timestamp": int(round(time.time() * 1000)),
            "grid_shape": self.__shape,
            "tick_duration": self.tick_duration,
            "team_members": team_members,
            "world_ID": self.world_id,
            "vis_settings": {
                "vis_bg_clr": self.__visualization_bg_clr,
                "vis_bg_img": self.__visualization_bg_img
            }
        }

        return state

    def __check_action_is_possible(self, agent_id, action_name, action_kwargs):
        # If the action_name is None, the agent idles
        if action_name is None:
            result = ActionResult(ActionResult.IDLE_ACTION, succeeded=True)
            return result

        # Check if the agent still exists (you would only get here if the agent is removed during this tick).
        if agent_id not in self.__registered_agents.keys():
            result = ActionResult(ActionResult.AGENT_WAS_REMOVED.replace("{AGENT_ID}", agent_id), succeeded=False)
            return result

        if action_name is None:  # If action is None, we send an action result that no action was given (and succeeded)
            result = ActionResult(ActionResult.NO_ACTION_GIVEN, succeeded=True)

        # action known, but agent not capable of performing it
        elif action_name in self.__all_actions.keys() and \
                action_name not in self.__registered_agents[agent_id].action_set:
            result = ActionResult(ActionResult.AGENT_NOT_CAPABLE, succeeded=False)

        # Check if action is known
        elif action_name in self.__all_actions.keys():
            # Get action class
            action_class = self.__all_actions[action_name]
            # Make instance of action
            action = action_class()
            # Check if action is possible, if so we can perform the action otherwise we send an ActionResult that it was
            # not possible.
            result = action.is_possible(self, agent_id, **action_kwargs)

        else:  # If the action is not known
            warnings.warn(f"The action with name {action_name} was not found when checking whether this action is "
                          f"possible to perform by agent {agent_id}.")
            result = ActionResult(ActionResult.UNKNOWN_ACTION, succeeded=False)

        return result

    def __perform_action(self, agent_id, action_name, action_kwargs):

        # Check if the action will succeed
        result = self.__check_action_is_possible(agent_id, action_name, action_kwargs)

        # If it will succeed, perform it.
        if result.succeeded:

            # If the action is None, nothing has to change in the world
            if action_name is None:
                return result

            # Get action class
            action_class = self.__all_actions[action_name]
            # Make instance of action
            action = action_class()
            # Apply world mutation
            result = action.mutate(self, agent_id, **action_kwargs)

            # Get agent's send_result function
            set_action_result = self.__registered_agents[agent_id].set_action_result_func
            # Send result of mutation to agent
            set_action_result(result)

            # Update the grid
            self.__update_agent_location(agent_id)

        # Whether the action succeeded or not, we return the result
        return result

    def __set_agent_busy(self, action_name, action_kwargs, agent_id):

        # Check if the action_name is None, in which case we simply idle for one tick
        if action_name is None:
            duration_in_ticks = 0

        else:  # action is not None

            # Get action class
            action_class = self.__all_actions[action_name]
            # Make instance of action
            action = action_class()

            # Obtain the duration of the action, defaults to the one of the action class if not in action_kwargs, and
            # otherwise that of Action
            duration_in_ticks = action.duration_in_ticks
            if "action_duration" in action_kwargs.keys():
                duration_in_ticks = action_kwargs["action_duration"]

            # Older kwarg name, raises deprecation warning
            if "duration_in_ticks" in action_kwargs.keys():
                warnings.warn("'duration_in_ticks' is deprecated for setting an action's duration; use "
                              "'action_duration'.", PendingDeprecationWarning)
                duration_in_ticks = action_kwargs["duration_in_ticks"]

        # The agent is now busy performing this action
        self.__registered_agents[agent_id]._set_agent_busy(curr_tick=self.current_nr_ticks,
                                                           action_duration=duration_in_ticks)

        # Set the action and result in the agent so we know where the agent is busy with. In addition this is appended
        # to its properties so others know what agent did)
        self.__registered_agents[agent_id]._set_current_action(action_name=action_name, action_args=action_kwargs)

    def __update_agent_location(self, agent_id):
        # Get current location of the agent
        loc = self.__registered_agents[agent_id].location
        # Check if that spot in our list that represents the grid, is None or a list of other objects
        if self.__grid[loc[1], loc[0]] is not None:  # If not None, we append the agent id to it
            self.__grid[loc[1], loc[0]].append(agent_id)
        else:  # if none, we make a new list with the agent id in it.
            self.__grid[loc[1], loc[0]] = [agent_id]

        # Update the Agent Avatar's location as well
        self.__registered_agents[agent_id].location = loc

    def __update_obj_location(self, obj_id):
        loc = self.__environment_objects[obj_id].location
        if self.__grid[loc[1], loc[0]] is not None:
            self.__grid[loc[1], loc[0]].append(obj_id)
        else:
            self.__grid[loc[1], loc[0]] = [obj_id]

    def __warn(self, warn_str):
        return f"[@{self.__current_nr_ticks}] {warn_str}"


    @property
    def registered_agents(self):
        """Dict: Dictionary of all registered agents, keys are the IDs, values are the registered objects. """
        return self.__registered_agents

    @property
    def environment_objects(self):
        """Dict: Dictionary of all non-agent environment objects, keys are the IDs, values are the registered objects.
        """
        return self.__environment_objects

    @property
    def is_done(self):
        """Bool: Boolean that indicates whether the GridWorld is done: either stopped by the user or due to the goal
         having been achieved."""
        return self.__is_done

    @property
    def current_nr_ticks(self):
        """Int: Current tick at which the gridworld is. """
        return self.__current_nr_ticks

    @property
    def grid(self):
        """Numpy 2D array: Numpy array of shape x by y. Each grid[x,y] location contains a list with all
        object IDs of the objects at that location"""
        return self.__grid

    @property
    def shape(self):
        """list: [x,y] shape of the grid """
        return self.__shape

    @property
    def simulation_goal(self):
        """WorldGoal: The world goal of type WorldGoal, or a class that extends WorldGoal """
        return self.__simulation_goal

    @property
    def tick_duration(self):
        """float: the desired duration of one tick. The real tick_duration might be longer due to a large amount of
         processing that needs to be done each tick by one or multiple agents. """
        return self.__tick_duration



import csv
import datetime
import os


class GridWorldLogger:
    LOG_ON_LAST_TICK = "log_last_tick"
    LOG_ON_FIRST_TICK = "log_first_tick"
    LOG_ON_GOAL_REACHED = "log_on_reached_goal"

    def __init__(self, log_strategy=1, save_path="/logs", file_name="", file_extension=".csv", delimiter=";"):
        self.__log_strategy = log_strategy
        self.__save_path = save_path
        self.__file_name_prefix = file_name
        self.__file_extension = file_extension
        self.__delimiter = delimiter

        # Create the file name
        current_time = datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.__file_name = f"{file_name}_{current_time}{file_extension}"

        self.__last_logged_tick = -1  # to track when we logged last
        self.__columns = []  # place holder for the columns in our data file
        self.__prev_goal_status = {}  # to track if a goal was accomplished since last call

    def log(self, grid_world, agent_data):
        return {}

    def _grid_world_log(self, grid_world, agent_data, last_tick=False, goal_status=None):
        if not self._needs_to_log(grid_world, last_tick, goal_status):
            return

        data = self.log(grid_world, agent_data)
        if data is None or data == {}:
            return

        self.__check_data(data)

        self.__write_data(data, grid_world.current_nr_ticks)

    def _needs_to_log(self, grid_world, last_tick, goal_status):
        current_tick = grid_world.current_nr_ticks

        # If the strategy is a tick frequency, check if enough ticks have passed
        if isinstance(self.__log_strategy, int):
            # the current nr ticks minus the tick we last logged should be smaller or equal to our frequency
            to_log = (current_tick - self.__last_logged_tick) >= self.__log_strategy
            if to_log:
                self.__last_logged_tick = current_tick

        # if the strategy is to log at the first tick, we do so if the current tick is zero
        elif self.__log_strategy == self.LOG_ON_FIRST_TICK:
            to_log = current_tick == 0

        # if the strategy is to log whenever one of the goals was reached
        elif self.__log_strategy == self.LOG_ON_GOAL_REACHED:
            to_log = False
            # we loop over all goals and see it's status became True whereas it was the previous time False
            for goal, status in goal_status.items():
                if goal in self.__prev_goal_status.keys():
                    if status and not self.__prev_goal_status[goal]:
                        to_log = True
                        break
            self.__prev_goal_status = goal_status.copy()

        # If we log on the last tick, only if the GridWorld says it is done
        elif self.__log_strategy == self.LOG_ON_LAST_TICK:
            to_log = last_tick

        # If the strategy is not found, we return an exception
        else:
            raise Exception(f"The log strategy {self.__log_strategy} is not recognized. Should be an integer or one of"
                            f"the GridWorld.ON_LOG_<...> values.")

        return to_log

    def _set_world_nr(self, world_nr):
        # Set the world number
        self.__world_nr = world_nr

        # Create the total file name path based on the world number (set by the WorldBuilder on logger creation)
        self.__save_path = f"{self.__save_path}{os.sep}world_{world_nr}"

        # Create the directory if not given
        if not os.path.exists(self.__save_path):
            os.makedirs(self.__save_path)

        self.__file_name = f"{self.__save_path}{os.sep}{self.__file_name}"

    def __check_data(self, data):
        if isinstance(data, dict):

            # Check if the data contains new columns, if so raise an exception that we cannot add columns on the fly
            if len(self.__columns) > 0:
                new_columns = set(data.keys()) - set(self.__columns)
                if len(new_columns) > 0:
                    raise Exception(f"Cannot append columns to the log file when we already logged with different "
                                    f"columns. THe following columns are new; {list(new_columns)}")

            return True
        else:
            raise Exception(f"The data in this {self.__class__} should be a dictionary.")

    def __write_data(self, data, tick_nr):

        # We always include the world number and the tick number
        if "world_nr" not in data.keys():
            data["world_nr"] = self.__world_nr
        if "tick_nr" not in data.keys():
            data["tick_nr"] = tick_nr

        # Check if we have columns to write to, this will be the order in which we write them
        if len(self.__columns) == 0:
            # Then we set the keys as column names
            self.__columns = list(data.keys())

        # Write the data to the file, create it when it does not exist and in that case write the columns as well
        if not os.path.isfile(self.__file_name):
            mode = "w+"
            write_columns = True
        else:
            mode = "a"
            write_columns = False

        with open(self.__file_name, mode=mode,  newline='') as data_file:
            csv_writer = csv.DictWriter(data_file, delimiter=self.__delimiter, quotechar='"',
                                        quoting=csv.QUOTE_MINIMAL, fieldnames=self.__columns)

            # Write columns if we need to
            if write_columns:
                csv_writer.writeheader()

            csv_writer.writerow(data)



import itertools
from collections import OrderedDict
from itertools import product

from matrx import WorldBuilder, utils
import numpy as np
from matrx.actions import MoveNorth, OpenDoorAction, CloseDoorAction
from matrx.actions.move_actions import MoveEast, MoveSouth, MoveWest
from matrx.agents import AgentBrain, HumanAgentBrain, SenseCapability
from matrx.cases.bw4t.bw4t_agents import BlockWorldAgent
from matrx.cases.bw4t.bw4t_objects import CollectBlock, SignalBlock
from matrx.grid_world import GridWorld, DropObject, GrabObject, AgentBody
from matrx.objects import EnvObject, SquareBlock
from matrx.world_builder import RandomProperty
from matrx.goals import WorldGoal, CollectionGoal

# Some general settings
tick_duration = 1 / 60  # 60fps if achievable
random_seed = 1
verbose = False
key_action_map = {  # For the human agents
    'w': MoveNorth.__name__,
    'd': MoveEast.__name__,
    's': MoveSouth.__name__,
    'a': MoveWest.__name__,
    'q': GrabObject.__name__,
    'e': DropObject.__name__,
    'r': OpenDoorAction.__name__,
    'f': CloseDoorAction.__name__,
}


def get_room_loc(room_nr):
    row = np.floor(room_nr / 3)
    column = room_nr % 3

    # x is: +1 for the edge, +edge hallway, +room width * column nr
    room_x = int(1 + 3 + (7 * column))

    # y is: +1 for the edge, +hallway space * (nr row + 1 for the top hallway), +row * room height
    room_y = int(1 + 3 * (row + 1) + row * 7)

    # door location is always center top
    door_x = room_x + int(np.floor(7 / 2))
    door_y = room_y

    return (room_x, room_y), (door_x, door_y)


def add_blocks(builder, room_locations, block_colours):
    for room_name, locations in room_locations.items():
        for loc in locations:
            # Get the block's name
            name = f"Block in {room_name}"

            # Get the probability for adding a block so we get the on average the requested number of blocks per room
            prob = min(1.0, 4 / len(locations))

            # Create a MATRX random property of color so each block varies per created world.
            # This random property is used to obtain a certain value (a color) for a certain object property
            # (visualize_colour) each time a new world is created from this builder.
            colour_property = RandomProperty(values=block_colours)

            # Add the block; a regular CollectibleObject as denoted by the given 'callable_class' which the
            # builder will use to create the object. In addition to setting MATRX properties, we also
            # provide a `is_block` boolean as custom property so we can identify this as a collectible
            # block.
            builder.add_object_prospect(loc, name, callable_class=CollectBlock, probability=prob,
                                        visualize_colour=colour_property)


def add_drop_off_zone(builder, world_size, block_colours, nr_blocks_to_collect):
    # First we calculate the top left coordinates of the room that contains our drop off zone
    x = 7
    y = 34
    door_x = 14
    door_y = y

    # Add the drop off room, with a door at its top
    builder.add_room(top_left_location=(x, y), width=15, height=7, name="Drop_off",
                     door_locations=[(door_x, door_y)])

    # Get all the locations INSIDE this room, again using a handy builder method.
    locs = utils.get_room_locations(room_top_left=(x, y), room_width=15, room_height=7)

    # Since we want each created world from our builder to have a different set of blocks to collect, we make a random
    # property that samples from all possible orderings of block colours (with duplicates). There is a handy class
    # method in `utils` that does this for us!
    possibilities = [{"visualization_colour": c} for c in block_colours]
    rp_order = CollectionGoal.get_random_order_property(possibilities, length=nr_blocks_to_collect,
                                               with_duplicates=True)

    # Create a collection goal. A collection goal in MATRX consists of the actual `CollectionGoal` that is linked to a
    # certain area that functions as the 'drop zone' and an invisible object containing a list of properties that
    # describe the to be collected objects. The `WorldBuilder` offers a handy method to create such goal, area and
    # invisible object. It requires the locations of this 'drop zone' and the list of properties describing the to be
    # collected objects. This list will become an actual property of our invisible object, thus it can also be a
    # RandomProperty instance! This enables us to vary the required order every time a new world is created. We just
    # made this property, so we just pass it through. We also set the colour and the opacity of the 'drop zone'.
    drop_zone_name = "Drop zone"
    builder.add_collection_goal(drop_zone_name, locs, rp_order, in_order=True,
                                collection_area_colour="#c87800",
                                collection_area_opacity=0.5, overwrite_goals=True)

    # Add our signal block that adapt itself to the then generated blocks to be collected.
    loc = (1, world_size[1] - 1 - nr_blocks_to_collect)
    for rank in range(nr_blocks_to_collect):
        builder.add_object(loc, name="Signal block", callable_class=SignalBlock, drop_zone_name=drop_zone_name,
                           rank=rank)
        loc = (loc[0], loc[1] + 1)


def add_agents(builder, block_sense_range, other_sense_range, memorize_for_ticks):
    # Create the agents sense capability. This is a circular range around the agent that denotes what it can perceive.
    # Here, we define that the agent cannot see other agent's their bodies, they can see square blocks with their own
    # range and see all other objects (doors, walls, etc.) with another range.
    sense_capability = SenseCapability({AgentBody: 0,
                                        CollectBlock: block_sense_range,
                                        None: other_sense_range})

    # Now we add our agents as part of the same team
    team_name = "Team Awesome"

    # We add 1 Human Agent; which is an agent you can control yourself. We first create the brain of the agent, and then
    # add it to our builder.
    brain = HumanAgentBrain(max_carry_objects=1, grab_range=0, drop_range=0, fov_occlusion=True,
                            memorize_for_ticks=memorize_for_ticks)
    loc = (1, 1)
    builder.add_human_agent(loc, brain, team=team_name, name="Human", key_action_map=key_action_map,
                            sense_capability=sense_capability)

    # We add 2 additional Autonomous Agents; an agent that does its thing without needing your input. Again, we create
    # its brain and add it to our builder. Since we provide the same team name, these agents will be in the same team as
    # the Human Agent.
    loc = (2, 1)
    brain = BlockWorldAgent()
    builder.add_agent(loc, brain, team=team_name, name=f"Agent Smith #1", sense_capability=sense_capability)
    loc = (3, 1)
    brain = BlockWorldAgent()
    builder.add_agent(loc, brain, team=team_name, name=f"Agent Smith #2", sense_capability=sense_capability)


def add_rooms(builder):
    room_locations = {}
    for room_nr in range(9):
        room_top_left, door_loc = get_room_loc(room_nr)

        # Add the room
        room_name = f"room_{room_nr}"
        builder.add_room(top_left_location=room_top_left, width=7, height=7, name=room_name,
                         door_locations=[door_loc], wall_visualize_colour="#8a8a8a",
                         with_area_tiles=True, area_visualize_colour="#dbdbdb", area_visualize_opacity=0.1)

        # Find all inner room locations where we allow objects (making sure that the location behind to door is free)
        room_locations[room_name] = utils.get_room_locations(room_top_left, 7, 7)

    return room_locations


def create_builder():
    # Some BW4T settings
    block_colours = ['#ff0000', '#ffffff', '#ffff00', '#0000ff', '#00ff00', '#ff00ff']
    block_sense_range = 10  # the range with which agents detect blocks
    other_sense_range = np.inf  # the range with which agents detect other objects (walls, doors, etc.)
    memorize_for_ticks = (10 / tick_duration)  # we want to memorize states for (seconds / tick_duration ticks) ticks

    # Set numpy's random generator
    np.random.seed(random_seed)

    # The world size, with plenty of space for agents to move between rooms
    world_size = (30, 44)

    # Create our world builder
    builder = WorldBuilder(shape=world_size, tick_duration=tick_duration, random_seed=random_seed, run_matrx_api=True,
                           run_matrx_visualizer=True, verbose=verbose, visualization_bg_clr="#f0f0f0",
                           visualization_bg_img="")

    # Add the world bounds (not needed, as agents cannot 'walk off' the grid, but for visual effect)
    builder.add_room(top_left_location=(0, 0), width=world_size[0], height=world_size[1], name="world_bounds")

    # Create the rooms
    room_locations = add_rooms(builder)

    # Add the blocks the agents need to collect, we do so probabilistically so each world will contain different blocks
    add_blocks(builder, room_locations, block_colours)

    # Create the drop-off zones, this includes generating the random colour/shape combinations to collect.
    add_drop_off_zone(builder, world_size, block_colours, nr_blocks_to_collect=2)

    # Add the agents and human agents to the top row of the world
    add_agents(builder, block_sense_range, other_sense_range, memorize_for_ticks)

    # Return the builder
    return builder


def run(nr_of_worlds=1):
    # Create our world builder
    builder = create_builder()

    # Start overarching MATRX scripts and threads, such as the api and/or visualizer if requested. Here we also link our
    # own media resource folder with MATRX.
    # media_folder = os.path.join(os.path.dirname(__file__), "media")
    builder.startup()

    for world in builder.worlds(nr_of_worlds=nr_of_worlds):
        world.run(builder.api_info)

    builder.stop()



from matrx.actions.action import Action, ActionResult
from matrx.actions.door_actions import OpenDoorAction, OpenDoorActionResult, CloseDoorAction, CloseDoorActionResult
from matrx.actions.move_actions import MoveActionResult, MoveNorth, MoveNorthEast, MoveSouthEast, MoveSouth, \
    MoveSouthWest, MoveWest, MoveNorthWest, MoveEast
from matrx.actions.object_actions import RemoveObject, RemoveObjectResult, GrabObject, GrabObjectResult, \
    DropObject, DropObjectResult



"""
    Author:         Aaron MacDonald
    Date:           June 14, 2007

    Description:    An implementation of the precise permissive field
                    of view algorithm for use in tile-based games.
                    Based on the algorithm presented at
                    http://roguebasin.roguelikedevelopment.org/index.php?title=Precise_Permissive_Field_of_View.
"""

import copy


def _field_of_view(start_x, start_y, map_width, map_height, radius, func_visit_tile, func_tile_blocked):
    """
    You are free to use or modify this code as long as this notice is
    included.
    This code is released without warranty.

    Determines which coordinates on a 2D grid are visible from a
    particular coordinate.

    startX, startY:         The (x, y) coordinate on the grid that
                            is the centre of view.

    mapWidth, mapHeight:    The maximum extents of the grid.  The
                            minimum extents are assumed to be both
                            zero.

    radius:                 How far the field of view may extend
                            in either direction along the x and y
                            axis.

    funcVisitTile:          User function that takes two integers
                            representing an (x, y) coordinate.  Is
                            used to "visit" visible coordinates.

    funcTileBlocked:        User function that takes two integers
                            representing an (x, y) coordinate.
                            Returns True if the coordinate blocks
                            sight to coordinates "behind" it.
    """

    visited = set()  # Keep track of what tiles have been visited so
    # that no tile will be visited twice.

    # Will always see the centre.
    func_visit_tile(start_x, start_y)
    visited.add((start_x, start_y))

    # Ge the dimensions of the actual field of view, making
    # sure not to go off the map or beyond the radius.

    if start_x < radius:
        min_extent_x = start_x
    else:
        min_extent_x = radius

    if map_width - start_x - 1 < radius:
        max_extent_x = map_width - start_x - 1
    else:
        max_extent_x = radius

    if start_y < radius:
        min_extent_y = start_y
    else:
        min_extent_y = radius

    if map_height - start_y - 1 < radius:
        max_extent_y = map_height - start_y - 1
    else:
        max_extent_y = radius

    # Northeast quadrant
    __check_quadrant(visited, start_x, start_y, 1, 1, max_extent_x, max_extent_y, func_visit_tile, func_tile_blocked)

    # Southeast quadrant
    __check_quadrant(visited, start_x, start_y, 1, -1,
                     max_extent_x, min_extent_y,
                     func_visit_tile, func_tile_blocked)

    # Southwest quadrant
    __check_quadrant(visited, start_x, start_y, -1, -1,
                     min_extent_x, min_extent_y,
                     func_visit_tile, func_tile_blocked)

    # Northwest quadrant
    __check_quadrant(visited, start_x, start_y, -1, 1,
                     min_extent_x, max_extent_y,
                     func_visit_tile, func_tile_blocked)


# -------------------------------------------------------------

class __Line(object):
    def __init__(self, xi, yi, xf, yf):
        self.xi = xi
        self.yi = yi
        self.xf = xf
        self.yf = yf

    dx = property(fget=lambda self: self.xf - self.xi)
    dy = property(fget=lambda self: self.yf - self.yi)

    def p_below(self, x, y):
        return self.relative_slope(x, y) > 0

    def p_below_or_collinear(self, x, y):
        return self.relative_slope(x, y) >= 0

    def p_above(self, x, y):
        return self.relative_slope(x, y) < 0

    def p_above_or_collinear(self, x, y):
        return self.relative_slope(x, y) <= 0

    def p_collinear(self, x, y):
        return self.relative_slope(x, y) == 0

    def line_collinear(self, line):
        return self.p_collinear(line.xi, line.yi) \
               and self.p_collinear(line.xf, line.yf)

    def relative_slope(self, x, y):
        return (self.dy * (self.xf - x)) \
               - (self.dx * (self.yf - y))


class __ViewBump:
    def __init__(self, x, y, parent):
        self.x = x
        self.y = y
        self.parent = parent


class __View:
    def __init__(self, shallow_line, steep_line):
        self.shallow_line = shallow_line
        self.steep_line = steep_line

        self.shallow_bump = None
        self.steep_bump = None


def __check_quadrant(visited, start_x, start_y, dx, dy,
                     extent_x, extent_y, func_visit_tile, func_tile_blocked):
    active_views = []

    shallow_line = __Line(0, 1, extent_x, 0)
    steep_line = __Line(1, 0, 0, extent_y)

    active_views.append(__View(shallow_line, steep_line))
    view_index = 0

    # Visit the tiles diagonally and going outwards
    #
    # .
    # .
    # .           .
    # 9        .
    # 5  8  .
    # 2  4  7
    # @  1  3  6  .  .  .
    max_i = extent_x + extent_y
    i = 1
    while i != max_i + 1 and len(active_views) > 0:
        if 0 > i - extent_x:
            start_j = 0
        else:
            start_j = i - extent_x

        if i < extent_y:
            max_j = i
        else:
            max_j = extent_y

        j = start_j
        while j != max_j + 1 and view_index < len(active_views):
            x = i - j
            y = j
            __visit_coord(visited, start_x, start_y, x, y, dx, dy,
                          view_index, active_views,
                          func_visit_tile, func_tile_blocked)

            j += 1

        i += 1


def __visit_coord(visited, start_x, start_y, x, y, dx, dy, view_index,
                  active_views, func_visit_tile, func_tile_blocked):
    # The top left and bottom right corners of the current coordinate.
    top_left = (x, y + 1)
    bottom_right = (x + 1, y)

    while view_index < len(active_views) \
            and active_views[view_index].steep_line.p_below_or_collinear(bottom_right[0], bottom_right[1]):
        # The current coordinate is above the current view and is
        # ignored.  The steeper fields may need it though.
        view_index += 1

    if view_index == len(active_views) \
            or active_views[view_index].shallow_line.p_above_or_collinear(top_left[0], top_left[1]):
        # Either the current coordinate is above all of the fields
        # or it is below all of the fields.
        return

    # It is now known that the current coordinate is between the steep
    # and shallow lines of the current view.

    is_blocked = False

    # The real quadrant coordinates
    real_x = x * dx
    real_y = y * dy

    if (start_x + real_x, start_y + real_y) not in visited:
        visited.add((start_x + real_x, start_y + real_y))
        func_visit_tile(start_x + real_x, start_y + real_y)
    """else:
        # Debugging
        print (start_x + real_x, start_y + real_y)"""

    is_blocked = func_tile_blocked(start_x + real_x, start_y + real_y)

    if not is_blocked:
        # The current coordinate does not block sight and therefore
        # has no effect on the view.
        return

    if active_views[view_index].shallow_line.p_above(bottom_right[0], bottom_right[1]) \
            and active_views[view_index].steep_line.p_below(top_left[0], top_left[1]):
        # The current coordinate is intersected by both lines in the
        # current view.  The view is completely blocked.
        del active_views[view_index]
    elif active_views[view_index].shallow_line.p_above(
            bottom_right[0], bottom_right[1]):
        # The current coordinate is intersected by the shallow line of
        # the current view.  The shallow line needs to be raised.
        __add_shallow_bump(top_left[0], top_left[1],
                           active_views, view_index)
        __check_view(active_views, view_index)
    elif active_views[view_index].steep_line.p_below(
            top_left[0], top_left[1]):
        # The current coordinate is intersected by the steep line of
        # the current view.  The steep line needs to be lowered.
        __add_steep_bump(bottom_right[0], bottom_right[1], active_views,
                         view_index)
        __check_view(active_views, view_index)
    else:
        # The current coordinate is completely between the two lines
        # of the current view.  Split the current view into two views
        # above and below the current coordinate.

        shallow_view_index = view_index
        view_index += 1
        steep_view_index = view_index

        active_views.insert(shallow_view_index, copy.deepcopy(active_views[shallow_view_index]))

        __add_steep_bump(bottom_right[0], bottom_right[1],
                         active_views, shallow_view_index)
        if not __check_view(active_views, shallow_view_index):
            view_index -= 1
            steep_view_index -= 1

        __add_shallow_bump(top_left[0], top_left[1], active_views,
                           steep_view_index)
        __check_view(active_views, steep_view_index)


def __add_shallow_bump(x, y, active_views, view_index):
    active_views[view_index].shallow_line.xf = x
    active_views[view_index].shallow_line.yf = y

    active_views[view_index].shallow_bump = __ViewBump(x, y, active_views[view_index].shallow_bump)

    cur_bump = active_views[view_index].steep_bump
    while cur_bump is not None:
        if active_views[view_index].shallow_line.p_above(cur_bump.x, cur_bump.y):
            active_views[view_index].shallow_line.xi = cur_bump.x
            active_views[view_index].shallow_line.yi = cur_bump.y

        cur_bump = cur_bump.parent


def __add_steep_bump(x, y, active_views, view_index):
    active_views[view_index].steep_line.xf = x
    active_views[view_index].steep_line.yf = y

    active_views[view_index].steep_bump = __ViewBump(x, y, active_views[view_index].steep_bump)

    cur_bump = active_views[view_index].shallow_bump
    while cur_bump is not None:
        if active_views[view_index].steep_line.p_below(cur_bump.x, cur_bump.y):
            active_views[view_index].steep_line.xi = cur_bump.x
            active_views[view_index].steep_line.yi = cur_bump.y

        cur_bump = cur_bump.parent


def __check_view(active_views, view_index):
    """
        Removes the view in active_views at index view_index if
            - The two lines are coolinear
            - The lines pass through either extremity
    """

    shallow_line = active_views[view_index].shallow_line
    steep_line = active_views[view_index].steep_line

    if shallow_line.line_collinear(steep_line) and (shallow_line.p_collinear(0, 1) or shallow_line.p_collinear(1, 0)):
        del active_views[view_index]
        return False
    else:
        return True



import copy

from matrx.messages.message import Message


class MessageManager:
    """ A manager inside the GirdWorld that tracks the received and send messages between agents and their teams.

        This provides several advantages:
        - an easier connection between MATRX Core, MATRX API, and the Front-end for
        messages (e.g. to differentiate between messages send between agents and to teams in the Front-end through a
        simple api call: 'get_team_messages').
        - an easy way to log communication (as the messages are easily obtained from a GridWorld instance, through some
        methods).
    """

    def __init__(self):
        # there are three types of messages
        self.global_messages = {}  # messages send to everyone
        self.team_messages = {}  # messages send to a team
        self.private_messages = {}  # messages send to individual agents

        self.preprocessed_messages = {}  # all types of messages above unpacked to the corresponding individual messages

        self.agents = None
        self.teams = None
        self.current_available_tick = 0

    def preprocess_messages(self, tick, messages, all_agent_ids, teams):
        """ Preprocess messages for sending, such that they can be understood by the GridWorld.

        For example: if the receiver=None, this means it must be sent to all agents. This function will process
         the receiver=None to a seperate message directed at every agent.

        Parameters
        ----------
        tick : int
            Current tick of the gridworld
        messages : dict
            All messages sent from the agent brains in the gridworld, and received via the api
        all_agent_ids : list
            IDs of all the agents
        teams : list
            ...

        -------
        """
        self.teams = teams
        self.agents = all_agent_ids

        # process every message
        for mssg in messages:

            # initialize a list for messages for this tick
            if tick not in self.preprocessed_messages:
                self.preprocessed_messages[tick] = []

            # check the message for validity
            MessageManager.__check_message(mssg, mssg.from_id)

            # decode the receiver_string into agent / team / global messages, save seperatly, and split into individual
            # messages understandable by the GridWorld
            self._decode_message_receiver(mssg, all_agent_ids, teams, tick)


    def _decode_message_receiver(self, mssg, all_agent_ids, teams, tick):
        """ Processes messages directed at other agents / teams / everyone.

        These messsage types are called private, team, and global messages.
        Messages of each type are saved for every tick separately, as well as a list with all preprocessed messages
        suitable for sending by the GridWorld.

        Possible formats for mssg.to_id
        "agent1"                  = private message to agent1 + team "agent1" if it exists
        ["agent1", "agent2"]      = 2 private messages + team messages if likewise named teams exist
        "team2"                   = team message sent to everyone in that team
        '["agent3", "team4"]'     = team + agent message. Provided as a string via the api
        None                      = global message send to everyone

        Parameters
        ----------
        mssg
            The original mssg object sent by the agent, or received via the api
        all_agent_ids
            List with IDs of all agents
        teams
            Dict with all team names (keys), and a list with all agent IDs in that team (value)
        tick
            Current tick of the gridworld
        """
        all_ids_except_me = [agent_id for agent_id in all_agent_ids if agent_id != mssg.from_id]

        # make sure when we copy this mssg to the global, private or team message groups, that we keep the potentially
        # custom message class and its custom properties. Also make sure that the message has a unique ID
        new_mssg = copy.copy(mssg).regen_id()

        # if the receiver is None, it is a global message which has to be sent to everyone
        if mssg.to_id is None:  # global message
            # create a list in memory for global messages for this tick
            if tick not in self.global_messages:
                self.global_messages[tick] = []

            # save in global
            global_message = self.copy_message(mssg=mssg, from_id=mssg.from_id, to_id="global")
            self.global_messages[tick].append(global_message)

            # split in sub mssgs
            for to_id in all_ids_except_me.copy():
                # create message
                new_message = self.copy_message(mssg=mssg, from_id=mssg.from_id, to_id=to_id)

                # save in prepr
                self.preprocessed_messages[tick].append(new_message)  # all messages above combined

        # if it is a list, decode every receiver_id in that list again
        elif isinstance(mssg.to_id, list):
            for receiver_id in mssg.to_id:
                # create message
                new_message = self.copy_message(mssg=mssg, from_id=mssg.from_id, to_id=receiver_id)

                self._decode_message_receiver(new_message, all_agent_ids, teams, tick)

        # a string might be: a list encoded as a string, a team, or an agent_id.
        elif isinstance(mssg.to_id, str):
            is_team_message = False

            try:
                # check if it is a list encoded as a string (sent via api)
                to_ids = eval(mssg.to_id)

                # create a new message addressed to this list of IDs, and reprocess
                new_message = self.copy_message(mssg=mssg, from_id=mssg.from_id, to_id=to_ids)
                self._decode_message_receiver(new_message, all_agent_ids, teams, tick)
            except:
                pass

            # Check if mssg_to is a team name. Note: an agent can only send a message to a team if they are in it.
            if mssg.to_id in teams.keys() and mssg.from_id in teams[mssg.to_id]:
                # create a list in memory for this team for this tick
                if tick not in self.team_messages:
                    self.team_messages[tick] = {}
                if mssg.to_id not in self.team_messages[tick]:
                    self.team_messages[tick][mssg.to_id] = []

                is_team_message = True

                # save in team mssgs as a message for that specific team
                self.team_messages[tick][mssg.to_id].append(mssg)

                # split in sub mssgs for every agent in the team
                for to_id in teams[mssg.to_id]:
                    # create a message
                    new_message = self.copy_message(mssg=mssg, from_id=mssg.from_id, to_id=to_id)

                    # save in prepr
                    self.preprocessed_messages[tick].append(new_message)

            # check if it is an agent ID (as well)
            # If no team is set by the user, the agent is added to a new team with the same name as the agent's ID.
            # As such, a message can be targeted at a team and individual agent at the same time.
            if mssg.to_id in all_agent_ids:
                # create a list in memory for private messages for this tick
                if tick not in self.private_messages:
                    self.private_messages[tick] = []

                # save in private_messages mssgs
                self.private_messages[tick].append(mssg)

                # if the message was not already saved in the preprocessed list, save it there as well
                if not is_team_message:
                    self.preprocessed_messages[tick].append(mssg)

    @staticmethod
    def __check_message(mssg, this_agent_id):
        if not isinstance(mssg, Message):
            raise Exception(f"A message to {this_agent_id} is not, nor inherits from, the class {Message.__name__}."
                            f" This is required for agents to be able to send and receive them.")

    def fetch_chatrooms(self, agent_id=None):
        """ Fetch all the chatrooms, or only those which a specific agent can view.

        Parameters
        ----------
        agent_id : str (optional, default, None)
            ID of the agent for which to fetch all accessible chatrooms. if None, all chatrooms are returned.

        Returns
        -------
        chatrooms : dict
            A dictionary containing a list with all "private" chatrooms, all "teams" chatrooms, and a "global" key,
            accessible via likewise named keys.
        """

        chatrooms = {"private": [], "team": [], "global": None}

        # add agents with which can be conversed
        for agent_id in self.agents:
            if agent_id != agent_id:
                chatrooms['private'].append(agent_id)

        # add teams with which can be conversed
        for team_name, team_members in self.teams.items():
            # if a specific agent ID is provided, only add teams of which that agent is a member
            if agent_id is not None:
                if agent_id in team_members or agent_id == "god":
                    chatrooms['team'].append(team_name)
            # otherwise, all teams can be conversed with
            else:
                chatrooms['team'].append(team_name)

        return chatrooms

    def fetch_messages(self, tick_from, tick_to, agent_id=None):
        """ Fetch messages, optionally filtered by start tick and/or agent id.

        Parameters
        ----------
        tick_from : int
            All messages from `tick_from` onwards to (including) `tick_to` will be collected.
        tick_to : int
            All messages from `tick_from` onwards to (including) `tick_to` will be collected.
        agent_id : string (optional, default, None)
            Only messages received by or sent by this agent will be collected. If none, all messages are returned
            between `tick_from` to `tick_to`.

        Returns
        -------
        Messages : dict
            Dictionary containing a 'global', 'team', and 'private' subdictionary.
                Global messages: messages['global'][tick] = [list of messages]
                Team messages: messages['team'][tick][team] = [list of messages]
                Private messages: messages['private'][tick] = [list of messages]

        Examples
        ---------
        In an action, world goal, or somewhere else with access to the Gridworld, the function can be used as below.
        This example returns all messages between tick 0 and 10.

        >>> messages = grid_world.message_manager.fetch_messages(0, 10, None)


        """
        messages = {'global': {}, 'team': {}, 'private': {}}

        # loop through all requested ticks
        for t in range(tick_from, tick_to + 1):
            # initialize the message objects to return
            # messages['global'][t] = None
            # messages['team'][t] = None
            # messages['private'][t] = None

            # fetch any existing glolbal messages for this tick
            if t in self.global_messages:
                # make the messages JSON serializable and add
                messages['global'][t] = [mssg.to_json() for mssg in self.global_messages[t]]

            # fetch any team messages
            if t in self.team_messages:
                if t not in messages['team']:
                    messages['team'][t] = {}

                # fetch all team messages
                if agent_id is None:
                    # make them JSON serializable
                    messages['team'][t] = [mssg.to_json() for mssg in self.team_messages[t]]

                # fetch team messages of the team of which the agent is a member
                else:
                    for team in self.teams:
                        if (agent_id in team or agent_id == "god") and team in self.team_messages[t]:
                            # make the messages JSON serializable and add
                            messages['team'][t][team] = [mssg.to_json() for mssg in self.team_messages[t][team]]

            # fetch private messages
            if t in self.private_messages:
                # fetch all private messages
                if agent_id is None:
                    # make the messages JSON serializable and add
                    messages['private'][t] = [mssg.to_json() for mssg in self.private_messages[t]]

                # fetch private messages sent to or received by the specified agent
                else:
                    messages['private'][t] = []
                    for message in self.private_messages[t]:
                        # only private messages addressed to or sent by our agent are requested
                        if message.from_id == agent_id or message.to_id == agent_id or agent_id == "god":
                            # make the messages JSON serializable and add
                            messages['private'][t].append(message.to_json())

        return messages

    def copy_message(self, mssg, from_id, to_id):
        """ Copy a message while keeping the potentially custom message type and custom message properties.
        Global and team messages have to be subdivided into individual messages for each receiving agent.
        This function copies a message while paying attention to any custom message classes used and their custom
        properties, in addition to making sure the message has a unique message ID.

        Parameters
        ----------
        mssg : (Custom)Message
            original message instance. Can be the default Message() class, or a class that inherits from it.
        from_id : str
            the new sender ID.
        to_id : str
             the new receiver ID.

        Returns
        new_mssg : (Custom)Message
            the new message instance with the same class and custom properties as mssg, with the provided from_id
            and to_id.
        -------
        """
        # copy the original message so we are sure we have the correct message class
        new_mssg = copy.copy(mssg)#

        # make sure the new message has a unique ID
        new_mssg.regen_id()

        # set the new from and to ID
        new_mssg.from_id = from_id
        new_mssg.to_id = to_id

        return new_mssg



import numpy as np


class Capability:
    """ Denotes an capability of an agent.

    Base class for agent capabilities.

    Notes
    -----
    Currently only used for the
    :class:`matrx.agents.capabilities.capability.SenseCapability`. Might be
    extended in the future to include other type of capabilities.

    """

    def __init__(self):
        pass


class SenseCapability(Capability):
    """ Denotes what an agent can see within a certain range.

    An instance of this class describes to an agent what it can perceive within
    what ranges. It is used by a :class:`matrx.grid_world.GridWorld` instance
    to construct the agent's state.

    To limit agents to sense objects more granulary then the given object types
    allow for, you can extend classes with your own classes. For example, one
    can extend the `SquareBlock` class into a custom class called
    `MySquareBlock`. This allows you to specify these two types separately to
    an agent for perception.

    Examples
    --------
    An example of a SenseCapability that defines the perception of all agents
    separate from SquareBlock objects. No other objects are perceived.

    >>> from matrx.objects import AgentBody, SquareBlock
    >>> SenseCapability({AgentBody: 10, SquareBlock: 25})

    An example of a SenseCapability that sets the perception range of all
    objects.

    >>> SenseCapability({None: 25})

    An example of a SenseCapability that sets the range of perceiving all
    other agents to 25 but the perception of all other objects to 10.

    >>> from matrx.objects import AgentBody
    >>> SenseCapability({AgentBody: 25, None: 10})

    """

    def __init__(self, detectable_objects):
        """ Denotes what an agent can see within a certain range.

        Parameters
        ----------
        detectable_objects : dict
            A dictionary with as keys the class you wish to perceive, and as
            values the distance this object type can be perceived. The None
            key stands for all otherwise not specified object types.
        """

        super().__init__()
        self.__detectable_objects = {}
        for obj_type, sense_range in detectable_objects.items():
            if obj_type is None:
                # If the obj_type is none, we can detect all object types in that associated range who have not a
                # specific range set.
                self.__detectable_objects["*"] = sense_range
            else:
                self.__detectable_objects[obj_type] = sense_range

    def get_capabilities(self):
        """ Returns the sense capabilities.

        Returns
        -------
        sense_capabilities: dict
            A dictionary with as keys the object types and values the
            distances.

            The key None denotes all object types.
        """
        return self.__detectable_objects.copy()


def create_sense_capability(objects_to_perceive, range_to_perceive_them_in):
    """ Creates a sense capability based that denotes what object types can be perceived from what range.


    Parameters
    ----------
    objects_to_perceive : list
        Various types of objects that the agent can perceive. An empty list is interpreted as all objects being
        detectable with infinite range by the agent.

    range_to_perceive_them_in : list
        The range for the object in `objects_to_perceive` from which the agent can perceive that object type.
        So range_to_perceive_them_in[4] denotes the range from which the agent can detect the object
        at objects_to_perceive[4].

    """
    # Check if range and objects are the same length
    assert len(objects_to_perceive) == len(range_to_perceive_them_in)

    # Check if lists are empty, if so return a capability to see all at any range
    if len(objects_to_perceive) == 0:
        return SenseCapability({"*": np.inf})

    # Create sense dictionary
    sense_dict = {}
    for idx, obj_class in enumerate(objects_to_perceive):
        perceive_range = range_to_perceive_them_in[idx]
        if perceive_range is None:
            perceive_range = np.inf
        sense_dict[obj_class] = perceive_range

    sense_capability = SenseCapability(sense_dict)

    return sense_capability



from matrx.logger.logger import GridWorldLogger


class LogActions(GridWorldLogger):

    def __init__(self, save_path="", file_name_prefix="", file_extension=".csv", delimeter=";"):
        super().__init__(save_path=save_path, file_name=file_name_prefix, file_extension=file_extension,
                         delimiter=delimeter, log_strategy=1)

    def log(self, grid_world, agent_data):
        log_data = {}
        for agent_id, agent_body in grid_world.registered_agents.items():
            log_data[agent_id] = agent_body.current_action

        return log_data



from matrx.logger.log_agent_actions import LogActions
from matrx.logger.log_idle_agents import LogIdleAgents
from matrx.logger.log_tick import LogDuration
from matrx.logger.logger import GridWorldLogger



import copy
import json

import numpy as np
from matrx.actions import GrabObject, RemoveObject, OpenDoorAction, CloseDoorAction
from matrx.agents.agent_utils.state import State
from matrx.messages import Message


class AgentBrain:
    """ An artificial agent whose behaviour can be programmed to be, for example, (semi-)autonomous.
    """

    def __init__(self, memorize_for_ticks=None):
        """ Defines the behavior of an agent.

        This class is the place where all the decision logic of an agent is
        contained. This class together with the
        :class:`matrx.objects.agent_body.AgentBody` class represent a full agent.

        This agent brain simply selects a random action from the possible actions
        it can do.

        When you wish to create a new agent, this is the class you need
        to extend. In specific these are the functions you should override:

        * :meth:`matrx.agents.agent_brain.initialize`
            Called before a world starts running. Can be used to initialize
            variables that can only be initialized after the brain is connected to
            its body (which is done by the world).
        * :meth:`matrx.agents.agent_brain.filter_observations`
            Called before deciding on an action to allow detailed and agent
            specific filtering of the received world state.
        * :meth:`matrx.agents.agent_brain.decide_on_action`
            Called to decide on an action.
        * :meth:`matrx.agents.agent_brain.get_log_data`
            Called by data loggers to obtain data that should be logged from this
            agent internal reasoning.

        Attributes
        ----------
        action_set: [str, ...]
            List of actions this agent can perform.
        agent_id: str
            The unique identified of this agent's body in the world.
        agent_name: str
            The name of this agent.
        agent_properties: dict
            A dictionary of this agent's
            :class:`matrx.objects.agent_body.AgentBody` properties. With as keys
            the property name, and as value the property's value.

            These can be adjusted iff they are said to be adjustable (e.g. inside
            the attribute `keys_of_agent_writable_props`).
        keys_of_agent_writable_props: [str, ...]
            List of property names that this agent can adjust.
        messages_to_send: [Message, ...]
            List of messages this agent will send. Use the method
            :meth:`matrx.agents.agent_brain.AgentBrain.send_message` to append to
            this list.
        previous_action: str
            The name of the previous performed or attempted action.
        previous_action_result: ActionResult
            The :class:`matrx.actions.action.ActionResult` of the previously
            performed or attempted action.
        received_messages: [Message, ...]
            The list of received messages.
        rnd_gen: Random
            The random generator for this agent.
        rnd_seed: int
            The random seed with which this agent's `rnd_gen` was initialized. This
            seed is based on the master random seed given of the
            :class:`matrx.grid_world.GridWorld`.
        """
        # Class variables for tracking the past action and its result
        self.previous_action = None
        self.previous_action_result = None

        # A list of messages that may be filled by this agent, which is retrieved by the GridWorld and send towards the
        # appropriate agents.
        self.messages_to_send = []
        self.received_messages = []

        # Filled by the WorldFactory during self.factory_initialise()
        self.agent_id = None
        self.agent_name = None
        self.action_set = None  # list of action names (strings)
        self.sense_capability = None
        self.rnd_gen = None
        self.rnd_seed = None
        self.agent_properties = {}
        self.keys_of_agent_writable_props = []
        self.__memorize_for_ticks = memorize_for_ticks

        # The central state property (an extended dict with unique searching capabilities)
        self._state = None

    def initialize(self):
        """ To initialize an agent's brain.

        Method called at the start of a :class:`matrx.grid_world.GridWorld`.

        Here you can initialize everything you need for your agent to work
        since you can't do much in the constructor as the brain needs to be
        connected to a GridWorld first in most cases (e.g. to get an AgentID,
        its random generator, etc.)
        """

    def filter_observations(self, state):
        """ Filters the world state before deciding on an action.

        In this method you filter the received world state to only those
        properties and objects the agent is actually supposed to see.

        Currently the world returns ALL properties of ALL objects within a
        certain range(s), as specified by :
        class:`matrx.agents.capabilities.capability.SenseCapability`. But
        perhaps some objects are obscured because they are behind walls and
        this agent is not supposed to look through walls, or an agent is not
        able to see some properties of certain objects (e.g. colour).

        The adjusted world state that this function returns is directly fed to
        the agent's decide function. Furthermore, this returned world state is
        also fed through the MATRX API to any visualisations.

        Override this method when creating a new AgentBrain and you need to
        filter the world state further.

        Parameters
        ----------
        state: State
            A state description containing all perceived
            :class:`matrx.objects.env_object.EnvObject` and objects inheriting
            from this class within a certain range as defined by the
            :class:`matrx.agents.capabilities.capability.SenseCapability`.

            The keys are the unique identifiers, as values the properties of
            an object. See :class:`matrx.objects.env_object.EnvObject` for the
            kind of properties that are always included. It will also contain
            all properties for more specific objects that inherit from that
            class.

            Also includes a 'world' key that describes common information about
            the world (e.g. its size).

        Returns
        -------
        filtered_state : State
            A dictionary similar to `state` but describing the filtered state
            this agent perceives of the world.

        Notes
        -----
        A future version of MATRX will include handy utility function to make
        state filtering less of a hassle (e.g. to easily remove specific
        objects or properties, but also ray casting to remove objects behind
        other objects)

        """

        return state

    def decide_on_action(self, state):
        """ Contains the decision logic of the agent.

        This method determines what action the agent should perform. The
        :class:`matrx.grid_world.GridWorld` is responsible for deciding when
        an agent can perform an action, if so this method is called for each
        agent and fed with the world state from the `filter_observations`
        method.

        Two things need to be determined: action name and action arguments.

        The action is returned simply as the class name (as a string), and the
        action arguments as a dictionary with the keys the names of the keyword
        arguments. See the documentation of that action to find out which
        arguments.

        An argument that is always possible is that of action_duration, which
        denotes how many ticks this action should take and overrides the
        action duration set by the action implementation.

        Parameters
        ----------
        state : State
        A state description as given by the agent's
        :meth:`matrx.agents.agent_brain.AgentBrain.filter_observation` method.

        Returns
        -------
        action_name : str
            A string of the class name of an action that is also in the
            `action_set` class attribute. To ensure backwards compatibility
            we advise to use Action.__name__ where Action is the intended
            action.

        action_args : dict
            A dictionary with keys any action arguments and as values the
            actual argument values. If a required argument is missing an
            exception is raised, if an argument that is not used by that
            action a warning is printed. The argument applicable to all action
            is `action_duration`, which sets the number ticks the agent is put
            on hold by the GridWorld until the action's world mutation is
            actual performed and the agent can perform a new action (a value of
            0 is no wait, 1 means to wait 1 tick, etc.).

        Notes
        -----
        A future version of MATRX will include handy utility function to make
        agent decision-making less of a hassle. Think of a
        Belief-Desire-Intention (BDI) like structure, and perhaps even support
        for learning agents.
        """

        # Send a message to a random agent
        agents = []
        for obj_id, obj in state.items():

            if obj_id is "World":  # Skip the world properties
                continue

            classes = obj['class_inheritance']
            if AgentBrain.__name__ in classes:  # the object is an agent to which we can send our message
                agents.append(obj)
        selected_agent = self.rnd_gen.choice(agents)
        message_content = f"Hello, my name is {self.agent_name}"
        self.send_message(Message(content=message_content, from_id=self.agent_id, to_id=selected_agent['obj_id']))

        # Select a random action
        if self.action_set:
            action = self.rnd_gen.choice(self.action_set)
        else:
            action = None

        action_kwargs = {}

        if action == RemoveObject.__name__:
            action_kwargs['object_id'] = None

            # Get all perceived objects
            objects = list(state.keys())
            # Remove yourself from the object id list
            objects.remove(self.agent_properties["obj_id"])
            # Remove all objects that have 'agent' in the name (so we do not remove those, though agents without agent
            # in their name can still be removed).
            objects = [obj for obj in objects if 'agent' not in obj]
            # Choose a random object id (safety for when it is empty)
            if objects:
                object_id = self.rnd_gen.choice(objects)
                # Assign it
                action_kwargs['object_id'] = object_id
                # Select range as just enough to remove that object
                remove_range = int(np.ceil(np.linalg.norm(
                    np.array(state[object_id]['location']) - np.array(
                        state[self.agent_properties["obj_id"]]['location']))))
                # Safety for if object and agent are in the same location
                remove_range = max(remove_range, 0)
                # Assign it to the arguments list
                action_kwargs['remove_range'] = remove_range
            else:
                action_kwargs['object_id'] = None
                action_kwargs['remove_range'] = 0

        # if the agent randomly chose a grab action, choose a random object to pickup
        elif action == GrabObject.__name__:
            # Set grab range
            grab_range = 1

            # Set max amount of objects
            max_objects = 3

            # Assign it to the arguments list
            action_kwargs['grab_range'] = grab_range
            action_kwargs['max_objects'] = max_objects

            # Get all perceived objects
            objects = list(state.keys())

            # Remove yourself from the object id list
            objects.remove(self.agent_properties["obj_id"])
            # Remove all objects that have 'agent' in the name (so we do not remove those, though agents without agent
            # in their name can still be removed).
            objects = [obj for obj in objects if 'agent' not in obj]
            # Choose a random object id (safety for when it is empty)

            object_in_range = []
            for object_id in objects:
                # Select range as just enough to grab that object
                dist = int(np.ceil(np.linalg.norm(
                    np.array(state[object_id]['location']) - np.array(
                        state[self.agent_properties["obj_id"]]['location']))))
                if dist <= grab_range and state[object_id]["is_movable"]:
                    object_in_range.append(object_id)

            if object_in_range:
                # Select object
                object_id = self.rnd_gen.choice(object_in_range)

                # Assign it
                action_kwargs['object_id'] = object_id
            else:
                action_kwargs['object_id'] = None

        # if we randomly chose to do a open or close door action, find a door to open/close
        elif action == OpenDoorAction.__name__ or action == CloseDoorAction.__name__:

            action_kwargs['door_range'] = 1  # np.inf
            action_kwargs['object_id'] = None

            # Get all doors from the perceived objects
            objects = list(state.keys())
            doors = [obj for obj in objects
                     if 'class_inheritance' in state[obj] and state[obj]['class_inheritance'][0] == "Door"]

            # choose a random door
            if len(doors) > 0:
                action_kwargs['object_id'] = self.rnd_gen.choice(doors)

        return action, action_kwargs

    def get_log_data(self):
        """ Provides a dictionary of data for any Logger

        This method functions to relay data from an agent's decision logic (this AgentBrain class) through the GridWorld
        into a Logger. Here it can be further processed and stored.

        Returns
        -------
        data : dict
            A dictionary with keys identifiable names and the data as its value.
        """
        return {}

    def send_message(self, message):
        """  Sends a Message from this agent to others

        Method that allows you to construct a message that will be send to either a specified agent, a team of agents
        or all agents.

        Parameters
        ----------
        message: Message
            A message object that needs to be send. Should be of type Message. It's to_id can contain a single
            recipient, a list of recipients or None. If None, it is send to all other agents.

        """
        # Check if the message is a true message
        self.__check_message(message, self.agent_id)
        # Add the message to our list
        self.messages_to_send.append(message)

    def is_action_possible(self, action, action_kwargs):
        """ Checks if an action would be possible.

        This method can be called from the AgentBrain to check if a certain action is possible to perform with the
        current state of the GridWorld. It requires as input an action name and its arguments (if any), same as the
        decide_on_action method should return.

        This method does not guarantees that if the action is return by the brain it actually succeeds, as other agents
        may intervene.

        Parameters
        ----------
        action : str
            The name of an Action class.
        action_kwargs : dict
            A dictionary with keys any action arguments and as values the actual argument values.

        Returns
        -------
        succeeded : bool
            True if the action can be performed, False otherwise.
        action_results : ActionResult
            An ActionResult object containing the success or failure of the action, and (if failed) the reason why.

        """
        action_result = self.__callback_is_action_possible(self.agent_id, action, action_kwargs)

        return action_result.succeeded, action_result

    @property
    def state(self):
        return self._state

    @property
    def memorize_for_ticks(self):
        return self.__memorize_for_ticks


    def create_context_menu_for_other(self, agent_id_who_clicked, clicked_object_id, click_location):
        """ Generate options for a context menu for a specific object/location that a user NOT controlling this
        human agent opened.

        Thus: another human agent selected this agent, opened a context menu by right clicking on an object or location.
        This function is called. It should return actions, messages, or other info for what this agent can do relevant
        to that object / location. E.g. pick it up, move to it, display information on it, etc.

        Example usecase: tasking another agent that is not yourself, e.g. to move to a specific location.

        For the default MATRX visualization, the context menu is opened by right clicking on an object. This function
        should generate a list of options (actions, messages, or something else) which relate to that object or location.
        Each option is in the shape of a text shown in the context menu, and a message which is send to this agent if
        the user actually clicks that context menu option.

        Parameters
        ----------
        agent_id_who_clicked : str
            The ID of the (human) agent that selected this agent and requested for a context menu.
        clicked_object_id : str
            A string indicating the ID of an object. Is None if the user clicked on a background tile (which has no ID).
        click_location : list
            A list containing the [x,y] coordinates of the object on which the user right clicked.

        Returns
        -------
         context_menu : list
            A list containing context menu items. Each context menu item is a dict with a 'OptionText' key, which is
            the text shown in the menu for the option, and a 'Message' key, which is the message instance that is sent
            to this agent when the user clicks on the context menu option.
        """
        print("Context menu other")
        context_menu = []

        # Generate a context menu option for every action
        for action in self.action_set:
            context_menu.append({
                "OptionText": f"Do action: {action}",
                "Message": Message(content=action, from_id=clicked_object_id, to_id=self.agent_id)
            })
        return context_menu

    def _factory_initialise(self, agent_name, agent_id, action_set, sense_capability, agent_properties,
                            customizable_properties, rnd_seed, callback_is_action_possible):
        """ Private MATRX function.

        Initialization of the brain by the WorldBuilder.

        Called by the WorldFactory to initialise this agent with all required properties in addition with any custom
        properties. This also sets the random number generator with a seed generated based on the random seed of the
        world that is generated.

        Parameters
        ----------
        agent_name : str
            The name of the agent.
        agent_id : str
            The unique ID given by the world to this agent's avatar. So the agent knows what body is his.
        action_set : str
            The list of action names this agent is allowed to perform.
        sense_capability : SenseCapability
            The SenseCapability of the agent denoting what it can see withing what range.
        agent_properties : dict
            The dictionary of properties containing all mandatory and custom properties.
        customizable_properties : list
            A list of keys in agent_properties that this agent is allowed to change.
        rnd_seed : int
            The random seed used to set the random number generator self.rng
        callback_is_action_possible : callable
            A callback to a GridWorld method that can check if an action is possible.

        """

        # The name of the agent with which it is also known in the world
        self.agent_name = agent_name

        # The id of the agent
        self.agent_id = agent_id

        # The names of the actions this agent is allowed to perform
        self.action_set = action_set

        # Setting the random seed and rng
        self.rnd_seed = rnd_seed
        self._set_rnd_seed(seed=rnd_seed)

        # Initializing the State object
        self._init_state()

        # The SenseCapability of the agent; what it can see and within what range
        self.sense_capability = sense_capability

        # Contains the agent_properties
        self.agent_properties = agent_properties

        # Specifies the keys of properties in self.agent_properties which can  be changed by this Agent in this file. If
        # it is not writable, it can only be  updated through performing an action which updates that property (done by
        # the environment).
        # NOTE: Changing which properties are writable cannot be done during runtime! Only when adding it to the world.
        self.keys_of_agent_writable_props = customizable_properties

        # A callback to the GridWorld instance that can check whether any action (with its arguments) will succeed and
        # if not why not (in the form of an ActionResult).
        self.__callback_is_action_possible = callback_is_action_possible

    def _get_action(self, state, agent_properties, agent_id):
        """ Private MATRX function

        The function the environment calls. The environment receives this function object and calls it when it is time
        for this agent to select an action.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        state_dict: dict
            A state description containing all properties of EnvObject that are within a certain range as defined by
            self.sense_capability. It is a list of properties in a dictionary
        agent_properties: dict
            The properties of the agent, which might have been changed by the environment as a result of actions of
            this or other agents.
        agent_id: str
            the ID of this agent

        Returns
        -------
         filtered_state : dict
            The filtered state of this agent
        agent_properties : dict
            the agent properties which the agent might have changed,
        action : str
            an action string, which is the class name of one of the actions in the Action package.
        action_kwargs : dict
            Keyword arguments for the action

        """
        # Process any properties of this agent which were updated in the environment as a result of actions
        self.agent_properties = agent_properties

        # Update the state property of an agent with the GridWorld's state dictionary
        self._state.state_update(state)

        # Call the filter method to filter the observation
        self._state = self.filter_observations(self._state)
        if isinstance(self._state, dict):
            raise ValueError(f"The filter_observation function of "
                             f"{self.agent_id} does not return a State "
                             f"object, but a dictionary. Please return "
                             f"self.state.")

        # Call the method that decides on an action
        action, action_kwargs = self.decide_on_action(self._state)

        # Store the action so in the next call the agent still knows what it did
        self.previous_action = action

        # Get the dictionary from the State object
        filtered_state = self._state.as_dict()

        # Return the filtered state, the (updated) properties, the intended actions and any keyword arguments for that
        # action if needed.
        return filtered_state, self.agent_properties, action, action_kwargs

    def _fetch_state(self, state_dict):
        self._state.state_update(state_dict)
        state = self.filter_observations(self._state)
        filtered_state_dict = state.as_dict()
        return filtered_state_dict

    def _get_log_data(self):
        return self.get_log_data()

    def _set_action_result(self, action_result):
        """ A function that the environment calls (similarly as the self.get_action method) to set the action_result of the
        action this agent decided upon.

        Note, that the result is given AFTER the action is performed (if possible).
        Hence it is named the self.previous_action_result, as we can read its contents when we should decide on our
        NEXT action after the action whose result this is.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        action_result : ActionResult
            An object that inherits from ActionResult, containing a boolean whether the action succeeded and a string
            denoting the reason why it failed (if it did so).
        """
        self.previous_action_result = action_result

    def _set_rnd_seed(self, seed):
        """ The function that seeds this agent's random seed.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        seed : int
            The random seed this agent needs to be seeded with.
        """
        self.rnd_seed = seed
        self.rnd_gen = np.random.RandomState(self.rnd_seed)

    def _get_messages(self, all_agent_ids):
        """ Retrieves all message objects the agent has made in a tick, and returns those to the GridWorld for sending.
        It then removes all these messages!

        This method is called by the GridWorld.

        Note; This method should NOT be overridden!
        Parameters
        ----------
        all_agent_ids
            IDs of all agents

        Returns
        -------
            A list of message objects with a generic content, the sender (this agent's id) and optionally a
            receiver.
        """
        # # preproccesses messages such that they can be understand by the gridworld
        # preprocessed_messages = self.preprocess_messages(this_agent_id=self.agent_id, agent_ids=all_agent_ids,
        # messages=self.messages_to_send)

        send_messages = copy.copy(self.messages_to_send)

        # Remove all messages that need to be send, as we have send them now
        self.messages_to_send = []

        return send_messages

    def _set_messages(self, messages=None):
        """
        This method is called by the GridWorld.
        It sets all messages intended for this agent to a list that it can access and read.

        Note; This method should NOT be overridden!

        Parameters
        ----------
        messages : Dict (optional, default, None)
            A list of dictionaries that contain a 'from_id', 'to_id' and 'content. If messages is set to None (or no
            messages are used as input), only the previous messages are removed
        """

        # We empty all received messages as this is from the previous tick
        # self.received_messages = []

        # Loop through all messages and create a Message object out of the dictionaries.
        for mssg in messages:

            # Check if the message is of type Message (its content contains the actual message)
            AgentBrain.__check_message(mssg, self.agent_id)

            # Since each message is secretly wrapped inside a Message (as its content), we unpack its content and
            # set that as the actual received message.
            received_message = mssg.content

            # Add the message object to the received messages
            self.received_messages.append(received_message)

    def _init_state(self):
        self._state = State(memorize_for_ticks=self.memorize_for_ticks, own_id=self.agent_id)

    @staticmethod
    def __check_message(mssg, this_agent_id):
        if not isinstance(mssg, Message):
            raise Exception(f"A message to {this_agent_id} is not, nor inherits from, the class {Message.__name__}."
                            f" This is required for agents to be able to send and receive them.")



import os

from matrx.agents import PatrollingAgentBrain, HumanAgentBrain
from matrx.world_builder import WorldBuilder
from matrx.actions import *


def create_builder():
    tick_dur = 0.1
    factory = WorldBuilder(random_seed=1, shape=[14, 20], tick_duration=tick_dur, verbose=False, run_matrx_api=True,
                           run_matrx_visualizer=True, visualization_bg_clr="#f0f0f0", simulation_goal=int(10/tick_dur),
                           visualization_bg_img='/static/images/restaurant_bg.png')

    factory.add_room(top_left_location=[0, 0], width=14, height=20, name="world_bounds")

    n_agent = 5

    is_even = True
    is_reverse = True
    for x in range(1, n_agent + 1):
        if is_even:
            is_even = False
            if is_reverse:
                start = (x, 12)
                waypoints = [(x, 1), (x, 11)]
            else:
                start = (x, 1)
                waypoints = [(x, 11), (x, 1)]
        else:
            is_even = True
            is_reverse = False
            if is_reverse:
                start = (1, x)
                waypoints = [(11, x), (1, x)]
            else:
                start = (12, x)
                waypoints = [(1, x), (11, x)]

        navigating_agent = PatrollingAgentBrain(waypoints, move_speed=10)
        factory.add_agent(start, navigating_agent, name="navigate " + str(x), visualization_shape=2, has_menu=True,
                          is_traversable=False, visualize_when_busy=True)

    # add human agent
    key_action_map = {
        'w': MoveNorth.__name__,
        'd': MoveEast.__name__,
        's': MoveSouth.__name__,
        'a': MoveWest.__name__,
        'r': RemoveObject.__name__
    }
    factory.add_human_agent([5, 5], HumanAgentBrain(), name="human",
                            key_action_map=key_action_map, img_name="/static/images/transparent.png")

    factory.add_human_agent([6, 6], HumanAgentBrain(), name="human2",
                            key_action_map=key_action_map, img_name="/static/images/agent.gif", visualize_when_busy=True)

    # add a number of blocks to showcase the subtile functionality
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[0,0], is_traversable=True, visualize_colour="#c0392b", visualize_shape=2)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[0,1], is_traversable=True, visualize_colour="#9b59b6", visualize_shape=1)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[0,2], is_traversable=True, visualize_colour="#2980b9", visualize_shape=0)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[1,0], is_traversable=True, visualize_colour="#1abc9c", visualize_shape=2, visualize_size=0.75)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[1,1], is_traversable=True, img_name="/static/images/fire.gif")
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[1,2], is_traversable=True, visualize_colour="#27ae60", visualize_shape=0)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[2,0], is_traversable=True, visualize_colour="#f1c40f", visualize_shape=2, visualize_size=0.5)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[2,1], is_traversable=True, visualize_colour="#e67e22", visualize_shape=1)
    factory.add_object([6,7], "block", subtiles=[3,3], subtile_loc=[2,2], is_traversable=True, visualize_colour="#2e4053", visualize_shape=0)


    return factory


def run_vis_test(nr_of_worlds=1):
    builder = create_builder()

    # startup world-overarching MATRX scripts, such as the api and/or visualizer if requested
    media_folder = os.path.dirname(os.path.realpath(__file__)) # set our path for media files to our current folder
    builder.startup(media_folder=media_folder)

    # run each world
    for world in builder.worlds(nr_of_worlds=nr_of_worlds):
        world.run(builder.api_info)

    # stop MATRX scripts such as the api and visualizer (if used)
    builder.stop()
